"""
Readers for various time series file formats
"""
import fnmatch

import numpy as np


def read_dat_names(path):
    """
    Read time series names from an ASCII file with data arranged column-wise. The names are in the first non-commented
    row.

    Parameters
    ----------
    path : str
        File path

    Returns
    -------
    list
        Time series names
        
    Notes
    -----
    The names are extracted from the header row (the first non-commented row). The comment character is '#'. Time is
    assumed to be in the first column.
    """
    names = None
    with open(path) as f:
        for line in f:
            if not line.startswith("#"):
                # names in first row that is not a comment
                names = line.split()
                break
    
    if names is not None:
        # identify time key, check that there is only one
        timekeys = fnmatch.filter(names, '[Tt]ime*')
        if len(timekeys) < 1:
            raise KeyError(f"The file '{path}' does not contain a time vector")
        elif len(timekeys) > 1:
            raise KeyError(f"The file '{path}' contains duplicate time vectors")

    # skip the time array name assumed to be in the first column
    return names[1:]


def read_dat_data(path, ind=None):
    """
    Read time series arranged column wise on ascii formatted file.

    Parameters
    ----------
    path : str
        File path
    ind : list of integers, optional
        Defines which responses to include in returned array. Each of the indices in `ind` refer the sequence number
        of the reponses. Note that `0` is the index of the time array.

    Returns
    -------
    array
        Time and data

    Notes
    -----
    If ``ind`` is specified, time array is only included if `0` is included in the specified indices.

    If ``ind`` is specified, the response arrays (1-D) are stored in the returned 2-D array in the same order as
    specified. I.e. if ``ind=[0,10,2,3]``, then the response array with index `10` on .ts file is obtained from
    ``data[1,:]``.

    """
    with open(path, 'r') as f:
        # skip commented lines at beginning of file
        for line in f:
            # break at first uncommented line (which is the header row and should not be read here)
            if not line.startswith("#"):
                break
        
        # load data, skipping commented lines and the header row with keys (first uncommented line)
        data = np.loadtxt(f, skiprows=0, usecols=ind, unpack=True)

    return data


def write_dat_data(path, time: np.ndarray, data: dict, delim: str = "\t", skip_header: bool = False):
    """
    Write time series arranged column wise on ascii formatted file.

    Parameters
    ----------
    path : str
        File path
    time : array
        Time
    data : dict
        Time series data name vs. values
    delim : str, optional
        Column delimiter
    skip_header : bool, optional
        Skip header on file
    """
    with open(path, "w") as f:
        # write keys + time to ascii-file header
        if not skip_header:
            header = ["%15s%s" % (k, delim) for k, _ in data.items()]
            header.insert(0, "%15s%s" % ("time", delim))
            header += "\n"
            f.write("".join(header))

        # write data to ascii file
        out = ""
        for i in range(len(time)):
            # write time value
            out += "%15.7g%s" % (time[i], delim)

            # write data values column wise (position 1 refers to time series data,
            # index i refers to time step #)
            for _, arr in data.items():
                out += "%15.7g%s" % (arr[1][i], delim)

            out += "\n"

            # flush to file every 500th time step (for efficiency) and at the end
            if ((i != 0) and (i % 500 == 0)) or (i == len(time) - 1):
                f.write(out)
                out = ""



# encoding: utf8
"""
Basic functions for statistical inference.
"""
import numpy as np


def empirical_cdf(n, kind='mean'):
    """
    Empirical cumulative distribution function given a sample size.

    Parameters
    ----------
    n : int
        sample size
    kind : str, optional
        - 'mean':  ``i/(n+1)`` (aka. Weibull method)
        - 'median' ``(i-0.3)/(n+0.4)``
        - 'symmetrical': ``(i-0.5)/n``
        - 'beard': ``(i - 0.31)/(n + 0.38)`` (Jenkinson's/Beard's method)
        - 'gringorten': ``(i - 0.44)/(n + 0.12)`` (Gringorten's method)

    Returns
    -------
    array
        Empirical cumulative distribution function

    Notes
    -----
    Gumbel recommended the following quantile formulation ``Pi = i/(n+1)``.
    This formulation produces a symmetrical CDF in the sense that the same
    plotting positions will result from the data regardless of
    whether they are assembled in ascending or descending order.

    Jenkinson's/Beard's method is based on the "idea that a natural
    estimate for the plotting position is the median of its probability
    density distribution".

    A more sophisticated formulation ``Pi = (i-0.3)/(n+0.4)`` approximates the
    median of the distribution free estimate of the sample variate to about
    0.1% and, even for small values of `n`, produces parameter estimations
    comparable to the result obtained by maximum likelihood estimations
    (Bury, 1999, p43)

    The probability corresponding to the unbiased plotting position can be
    approximated by the Gringorten formula in the case of type 1 Extreme
    value distribution.

    References
    ----------
    1. `Plotting positions <http://en.wikipedia.org/wiki/Q%E2%80%93Q_plot>`_, About plotting positions
    """

    n = float(n)
    i = np.arange(n) + 1
    if kind == 'mean':
        f = i / (n + 1.)
    elif kind == 'median':
        f = (i - 0.3) / (n + 0.4)
    elif kind == 'symmetrical':
        f = (i - 0.5) / n
    elif kind == 'beard':
        f = (i - 0.31) / (n + 0.38)
    elif kind == 'gringorten':
        f = (i - 0.44) / (n + 0.12)
    else:
        raise ValueError("Distribution type must be either 'mean','median','symmetrical','beard','gringorten'")

    return f




#!/usr/bin/env python
# coding: utf-8
"""
Fatigue related corrections.
"""
import numpy as np


def goodman_haigh(cycles, uts):
    """
    Effective alternating stress using the Goodman-Haigh mean stress correction for fully reversed loading (R=-1).

    Parameters
    ----------
    cycles : np.ndarray or list
        Array of cycle ranges and mean values, shape `(n, 2)`. First column is cycle ranges, second column is cycle
        mean values.
    uts : float
        Material ultimate tensile strength, in same unit as cycle ranges and mean values.

    Returns
    -------
    corrected_ranges: np.ndarray
        Corrected stress ranges, shape: `(n,)` (1D array).

    Notes
    -----
    In materials science and fatigue, the Goodman relation is an equation used to quantify the interaction of mean and
    alternating stresses on the fatigue life of a material.

    A Goodman diagram,[1][2] sometimes called a Haigh diagram or a Haigh-Soderberg diagram,[3] is a graph of (linear)
    mean stress vs. (linear) alternating stress, showing when the material fails at some given number of cycles.

    A scatterplot of experimental data shown on such a plot can often be approximated by a parabola known as the Gerber
    line, which can in turn be (conservatively) approximated by a straight line called the Goodman line.

    Correcting the stress ranges like this can only be applied with an SN-curve with a stress ratio (R) of -1.

    References
    ----------
    1. Herbert J. Sutherland and John F. Mandell. "Optimized Goodman diagram for the analysis of fiberglass composites
       used in wind turbine blades"
    2. David Roylance. "Fatigue", Archived 2011-06-29 at the Wayback Machine.. 2001
    3. Tapany Udomphol. "Fatigue of metals". 2007.

    Examples
    --------

    >>> from qats.fatigue.rainflow import count_cycles
    >>> from qats.fatigue.corrections import goodman_haigh
    >>> # assuming a series has been established, and that uts is defined
    >>> cycles = count_cycles(series)
    >>> corrected_ranges = goodman_haigh(cycles[:, :2], uts)

    The array obtained is then a 1D array of same size as the first dimension of `cycles`:

    >>> cycles.shape
    (1350, 3)
    >>> corrected_ranges.shape
    (1350,)

    """
    # ensure array and assert 2d
    cycles = np.asarray(cycles)
    assert cycles.ndim == 2 and cycles.shape[1] == 2, \
        "Cycles must be specified as 2D array or shape (n, 2) (or: list of 2-tuples)"

    # unpack
    ranges, means = cycles.T

    # calculate effective alternating stress
    corrected_ranges = ranges * (uts / (uts - means))

    return corrected_ranges




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Provides :class:`TsDB` class.
"""
import copy
import fnmatch
import glob
import os
from collections import OrderedDict, defaultdict
from uuid import uuid4

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from .fatigue.rainflow import rebin as rebin_cycles
from .io.csv import read_data as read_csv_data
from .io.csv import read_names as read_csv_names
from .io.direct_access import (read_tda_data, read_tda_names, read_ts_data,
                               read_ts_names, write_ts_data)
from .io.other import read_dat_data, read_dat_names, write_dat_data
from .io.sima import read_ascii_data as read_sima_ascii_data
from .io.sima import read_bin_data as read_sima_bin_data
from .io.sima import read_names as read_sima_names
from .io.sima import read_sima_wind_names
from .io.sima_h5 import read_data as read_sima_h5_data
from .io.sima_h5 import read_names as read_sima_h5_names
from .io.sima_h5 import write_data as write_sima_h5_data
from .io.sintef_mat import read_data as read_mat_data
from .io.sintef_mat import read_names as read_mat_names
from .io.tdms import read_data as read_tdms_data
from .io.tdms import read_names as read_tdms_names
from .ts import TimeSeries

# todo: cross spectrum(scipy.signal.csd)
# todo: coherence (scipy.signal.coherence)


class TsDB(object):
    """
    A class for storage, processing and presentation of time series.

    Parameters
    ----------
    name : str, optional
        Database name.

    Attributes
    ----------
    name : str
        Database name.
    register : OrderedDict
        TimeSeries objects by unique time series id.
    register_parent : OrderedDict
        Parent file path by unique time series id.
    register_indices : OrderedDict
        Index of time series on parent file by unique time series id.
    register_keys : list
        Unique time series id.

    """

    def __init__(self, name=None):
        self.uuid = uuid4()
        self.name = name
        self.register = OrderedDict()           # dictionary of unique id and time series objects
        self.register_parent = OrderedDict()    # dictionary of unique id and parent name (source/file name)
        self.register_indices = OrderedDict()   # dictionary of unique id and the time series index on parent file
        self.register_keys = []     # register keys in the order the associated time series where loaded
        self._timekeys = dict()  # register of time keys (only relevant for .mat files)

    def __contains__(self, item):
        if isinstance(item, str):
            # item interpreted as time series name
            names = item
        elif isinstance(item, TimeSeries):
            names = item.fullname
        else:
            raise TypeError(f"Unable to check containment of type '{type(item)}' items. Item must be string or"
                            f"TimeSeries.")

        match = self.list(names=names, display=False)
        if len(match) > 0:
            return True
        else:
            return False

    def __eq__(self, other):
        return isinstance(other, TsDB) and other.uuid == self.uuid

    def __iter__(self):
        """
        Generator yielding time series in database

        Returns
        -------
        TimeSeries
            time series in database
        """
        for key in self.register_keys:
            yield self.get(name=key)

    def __len__(self):
        return self.n

    def __repr__(self):
        return f"<TsDB id='{self.uuid}'>"

    def __str__(self):
        _ = f"type: TsDB\n" \
            f"id : {self.uuid}\n" \
            f"name : {self.name}\n" \
            f"number of time series : {self.n}\n"
        return _

    @classmethod
    def fromfile(cls, filenames, read=False, verbose=False):
        """
        Create TsDB instance from one ore more files.

        Parameters
        ----------
        filenames: str, list or tuple
            File names including suffix. Wildcards can also be used.
        read: bool, optional
            If True, all time series are read from file and stored. The default is that they are read from file
            when requested by any of the `get*()` methods.
        verbose : bool, optional
            If True, print information to screen.

        Returns
        -------
        TsDB
            TsDB instance with files loaded.

        Notes
        -----
        The purpose of the classmethod is efficient initiation of a new class instance. For example:

        >>> from qats import TsDB
        >>> tsdb = TsDB.fromfile("mooring.ts")

        ... is equivalent to:

        >>> tsdb = TsDB("")
        >>> tsdb.load("mooring.ts")


        See Also
        --------
        load : Includes notes of relevance for this method.
        """
        tsdb = cls("")
        tsdb.load(filenames, read=read, verbose=verbose)
        return tsdb

    @property
    def common(self):
        """
        Common part of all keys (paths) in DB.

        Returns
        -------
        str
        """
        if self.n == 0:
            return ""
        elif self.n == 1:
            k = self.register_keys[0]
            return self._path_dirname(k)
        else:
            try:
                return os.path.commonpath(self.register_keys)
            except ValueError:
                # if paths contain both absolute and relative paths, the paths are on the different
                # drives or if paths is empty.
                return ""

    @property
    def n(self):
        """
        Number of time series in database

        Returns
        -------
        int : number of loaded time series

        """
        return len(self.register)

    @staticmethod
    def _check_time_arrays(container, **kwargs):
        """

        Parameters
        ----------
        container: dict
            Time series container obtained from `TsDB.getm()`.
        kwargs:
            Optional arguments passed to `TimeSeries.get()`

        Returns
        -------
        dict
            Attributes: 'is_common' (bool), 'dtg_defined' (bool), 'dtg_ref' (datetime or None), 'common' (tuple),
            'devations' (dict)', 'actions' (list), 'msg' (str).
        """
        # evaluate whether the time series have common time array (time steps and time windows are equal)
        _dtg_ref = []
        _dt = []
        _start_time = []
        _end_time = []

        for _, ts in container.items():
            _dtg_ref.append(ts.dtg_ref)
            _dt.append(ts.dt)
            _start_time.append(ts.start)
            _end_time.append(ts.end)

        # evaluate dtg_ref, etc.
        _same_dtg_ref = all(r == _dtg_ref[0] for r in _dtg_ref)
        dtg_defined = any(r is not None for r in _dtg_ref)
        dtg_ref = _dtg_ref[0] if (dtg_defined is True and _same_dtg_ref is True) else None

        # evaluate time step, start and end time
        _same_dt = ((max(_dt) - min(_dt)) == 0.)
        _same_start_time = ((max(_start_time) - min(_start_time)) == 0.)
        _same_end_time = ((max(_end_time) - min(_end_time)) == 0.)

        # recommended parameters for common time: latest start, earliest end, smallest avg. time step
        common_start = max(_start_time)
        common_end = min(_end_time)
        if common_start < common_end:
            common = common_start, common_end, min(_dt)
        else:
            # not possible to create common time array (earliest end time is before latest start time)
            common = None

        # evaluate and define deviations and recommended actions
        deviations = OrderedDict()

        if not _same_dt:
            deviations["dt"] = "mean time step varies from %.10g to %.10g" % (min(_dt), max(_dt))

        if not _same_start_time:
            deviations["start"] = "start time varies from %.7g to %.7g" % (min(_start_time), max(_start_time))

        if not _same_end_time:
            deviations["end"] = "end time varies from %.7g to %.7g" % (min(_end_time), max(_end_time))

        if dtg_defined and not _same_dtg_ref:
            '''
            If dtg_ref test criteria fails, none of the previous tests are relevant anymore. Therefore, the devations
            dict is reset here.
            '''
            deviations = OrderedDict()
            deviations["dtg_ref"] = "'dtg_ref' is defined for one or more of the time series, " \
                                    "but is not equal for all of them"

        # identify any deviations that are handled by kwargs
        handled = set()
        if "resample" in kwargs:
            if isinstance(kwargs.get("resample"), float):
                handled.add("dt")
            elif isinstance(kwargs.get("resample"), np.ndarray) or isinstance(kwargs.get("resample"), list):
                t_new = kwargs.get("resample")
                if np.allclose(np.min(np.diff(t_new)), np.max(np.diff(t_new))):
                    handled.add("dt")
                if t_new[0] >= common_start:
                    handled.add("start")
                if t_new[-1] >= common_end:
                    handled.add("end")
        if kwargs.get("twin") is not None:
            t_start, t_end = kwargs["twin"]
            if t_start >= common_start:
                handled.add("start")
            if t_end >= common_end:
                handled.add("end")

        # remove deviations handled by kwargs
        for h in handled:
            _ = deviations.pop(h, None)

        # evaluate if time array is common
        is_common = (len(deviations) == 0)

        # define actions
        actions = []
        msg = ""
        if not is_common:
            if "dt" in deviations:
                actions.append("Resample to constant time step by specifying `resample=%.10g` (or smaller)" % common[2])
            if "start" in deviations or "end" in deviations:
                actions.append("Crop time series to the common time window by specifying "
                               "`twin=(%.7g, %.7g)`" % (common[0], common[1]))
                actions.append("Resample to a common time array by specifying "
                               "`resample=np.arange(%.7g, %.7g, %.7g)`" % (common[0], common[1]+common[2], common[2]))
            if "dtg_ref" in deviations:
                actions.append("Ensure 'dtg_ref' is the same for all specified time series "
                               "-- see `TimeSeries.set_dtg_ref()`")

            # compile deviations and recommended actions to message string
            msg = "Time array is not common within specified parameters for specified time series.\n" \
                  "Deviations:\n"
            for dev, s in deviations.items():
                msg += "  * %s : %s\n" % (dev, s)
            if len(actions) > 0:
                msg += "Recommended actions (one or more may be necessary):\n"
                if common is not None:
                    for a in actions:
                        msg += "  * %s\n" % a
                else:
                    msg += "   (none - not possible to create common time array)\n"

        timecheck = dict(
            is_common=is_common,        # True or False
            # dtg info
            dtg_defined=dtg_defined,    # True or False
            dtg_ref=dtg_ref,            # datetime or None
            # recommended parameters for common time array
            common=common,              # tuple or None
            # descriptions of deviations and proposed actions
            deviations=deviations,      # dict
            actions=actions,            # list
            msg=msg,                    # str
        )

        return timecheck

    @staticmethod
    def _path_basename(key):
        """
        As os.path.basename, but does not split on '/' or '\\' if they are within square brackets.
        """
        if "[" in key:
            i = key.index("[")
            return os.path.basename(key[:i]) + key[i:]
        else:
            return os.path.basename(key)

    @staticmethod
    def _path_dirname(key):
        """
        As os.path.dirname, but does not split on '/' or '\\' if they are within square brackets.
        """
        if "[" in key:
            i = key.index("[")
            return os.path.dirname(key[:i]) + key[i:]
        else:
            return os.path.dirname(key)

    @staticmethod
    def _path_relpath(key, start=os.curdir):
        """
        As os.path.relpath, but does not split on '/' or '\\' if they are within square brackets.
        """
        if "[" in key:
            i = key.index("[")
            return os.path.relpath(key[:i], start) + key[i:]
        else:
            return os.path.relpath(key, start)

    @staticmethod
    def _reorder_namelist(namelist, names=None):
        """
        Re-order time series names to match order in which they were specified.

        Parameters
        ----------
        namelist: list
            List of time series names to re-order.
        names: list, optional
            Time series name (patterns)

        Returns
        -------
        list
            Re-ordered time series names

        Raises
        ------
        ValueError
            If `keep_order` is True, in combination with more than one key and more than one name.
        """

        # define function need for sorting
        def get_index(m, patterns):
            """Return index of matched time series names."""
            try:
                _ind = [fnmatch.fnmatch(m, pat) for pat in patterns].index(True)
            except ValueError:
                raise Exception("Unexpected error: could not find sorting index for key '%s'" % m)
            return _ind

        # re-order names to match specified order
        if names is not None and len(names) > 1:
            return sorted(namelist, key=lambda m: get_index(m, names))
        else:
            # names or names not sufficiently specified, there is nothing to do
            return namelist

    def _make_export_friendly_names(self, container, keep_basename=False):
        """
        Shorten keys by removing common part of it and replace path separators with underscore. Use only basename
        if specified.

        Parameters
        ----------
        container : dict
            Container with time series
        keep_basename : bool, optional
            Keep only time series names e.g. 'tension' in key 'C:\\data\\results.ts\\tension'. Default False.

        Returns
        -------
        dict
            Time series container with modified keys

        Raises
        ------
        KeyError
            If `keep_basename` = True and several timeseries have the same name

        Notes
        -----
        To keep info but avoid path confusion when loading the exported file later
            - keys shortened by only removing non-unique part
            - remove parent file extension from key
            - replace path separators in key with underscore
            - do not replace '.' elsewhere in the key e.g. directory name or filename containing '.'

        """
        new_container = OrderedDict()

        # force keeping basename if there is only 1 ts in the container. This to avoid totally removing the name
        if len(container) == 1:
            keep_basename = True

        if keep_basename:
            key_count = defaultdict(int)
        else:
            # common part of all selected keys
            common_key = os.path.commonpath([str(k) for k in container.keys()])

        for key, ts in container.items():
            if keep_basename:
                # shorten key by using only basename
                new_k = self._path_basename(key)
                key_count[new_k] += 1
                if key_count[new_k] > 1:
                    raise KeyError("Multiple keys with basename '%s' -- "
                                   "consider using ``basename=False``" % new_k)
            else:
                # make
                relkey = self._path_relpath(key, common_key)
                name = self._path_basename(relkey)
                new_k = "_".join([os.path.splitext(self._path_dirname(relkey))[0].replace(os.path.sep, "_"), name])

                # Remove leading underscores in the name
                if new_k.startswith("_"):
                    new_k = new_k[1:]

            # put in new container
            new_container[new_k] = ts

        return new_container

    def _read(self, keys, retkeys=None, store=True):
        """
        Read time series specified by absolute keys from file

        Parameters
        ----------
        keys : list
            Unique time series identifiers/keys/path
        retkeys : list, optional
            Unique identifiers/keys used in returned container. Useful if you want to have shorter keys
            (less the common db path) in the container returned. Note that full key is used in the db register.
        store : bool, optional
            Disable time series storage. Default is to store the time series objects first time it is read.

        Returns
        -------
        dict
            TimeSeries objects

        Notes
        -----
        Absolute keys are obtained using the list() method.

        """
        # handle retkeys
        if retkeys is not None:
            assert len(retkeys) == len(keys), "The number of 'retkeys' is different from the number of 'keys'."
        else:
            retkeys = copy.copy(keys)

        # correlation between full keys and specified keys for returned container
        keypairs = dict(zip(keys, retkeys))

        # initiate and pre-populate ordered dictionary (to keep order of keys)
        container = OrderedDict()
        for key in keys:
            reg = self.register[key]   # gives None if not previously read/stored
            if isinstance(reg, TimeSeries) or reg is None:
                container[keypairs[key]] = reg
            else:
                raise LookupError("Unexpected look-up error, key: %s" % key)

        # group keys by parent file
        keys_by_parent = defaultdict(list)
        for key in keys:
            # only interested in keys that are not already stored
            if container[keypairs[key]] is None:
                keys_by_parent[self.register_parent[key]].append(key)

        # read requested keys, file by file
        for parent, keys in keys_by_parent.items():
            names = [key.replace(parent, "").lstrip(os.path.sep) for key in keys]

            # extract parent file extension
            fext = os.path.splitext(parent)[-1]

            # indices of time series to be read
            indices = [0] + [self.register_indices[key] for key in keys]

            tslist = [None] * len(keys)

            if fext == '.ts':
                data = read_ts_data(parent, ind=indices)
                for i, name in enumerate(names):
                    tslist[i] = TimeSeries(name, data[0, :], data[i + 1, :], parent=parent)

            elif fext == '.tda':
                data = read_tda_data(parent, ind=indices)
                for i, name in enumerate(names):
                    tslist[i] = TimeSeries(name, data[0, :], data[i+1, :], parent=parent)

            elif fext == '.asc':
                data = read_sima_ascii_data(parent, ind=indices)
                for i, name in enumerate(names):
                    tslist[i] = TimeSeries(name, data[0, :], data[i+1, :], parent=parent)

            elif fext == '.bin':
                data = read_sima_bin_data(parent, ind=indices)
                for i, name in enumerate(names):
                    tslist[i] = TimeSeries(name, data[0, :], data[i+1, :], parent=parent)

            elif fext == '.dat':
                data = read_dat_data(parent, ind=indices)
                for i, name in enumerate(names):
                    tslist[i] = TimeSeries(name, data[0, :], data[i+1, :], parent=parent)

            elif fext == '.mat':
                _tk = self._timekeys[parent]
                data = read_mat_data(parent, [_tk, *names])
                for i, name in enumerate(names):
                    tslist[i] = TimeSeries(name, data[_tk], data[name], parent=parent)

            elif fext in ('.h5', '.hdf5'):
                data = read_sima_h5_data(parent, names=names)
                for i, name in enumerate(names):
                    timearr, arr = data[i]
                    tslist[i] = TimeSeries(name, timearr, arr, parent=parent)

            elif fext == '.csv':
                data = read_csv_data(parent, ind=indices)
                for i, name in enumerate(names):
                    tslist[i] = TimeSeries(name, data[0, :], data[i + 1, :], parent=parent)

            elif fext == '.tdms':
                data = read_tdms_data(parent, names=names)
                for i, name in enumerate(names):
                    timearr, arr = data[i]
                    tslist[i] = TimeSeries(name, timearr, arr, parent=parent)
            else:
                raise NotImplementedError("Invalid file type: %s (ext = %s)" % (parent, fext))

            # add to ordered dictionary (and store in db if specified)
            for key, ts in zip(keys, tslist):
                container[keypairs[key]] = ts

                if store:
                    self.register[key] = ts

        return container

    def add(self, ts):
        """
        Add new TimeSeries object to db

        Parameters
        ----------
        ts : TimeSeries
            added TimeSeries object

        Notes
        -----
        Key/identifier will be the name of the time series. If you want to change the key, just change the name before
        adding the TimeSeries to the db.

        """
        '''
        Note: 
        If only 'name' is used as the register key, one may encounter issues later (e.g. at export), since 
        os.path.commonpath will not accept a mix of absolute and relative paths. Therefore, the added timeseries is
        registered with a fictitious key constructed as follows: commonpath + ts.name
        '''
        if not isinstance(ts, TimeSeries):
            raise TypeError("expected TimeSeries instance, got: %s" % type(ts))

        key = os.path.join(self.common, ts.name)

        if key in self.register.keys():
            raise KeyError("The specified key is not unique: %s" % key)

        self.register[key] = ts
        self.register_parent[key] = None    # does not have a parent (file)
        self.register_indices[key] = None   # ... and therefore has no index (yet)
        self.register_keys.append(key)

    def clear(self, names=None, display=True):
        """
        Clear/remove time series from register

        Parameters
        ----------
        names : str or list or tuple, optional
            Name of time series to remove from database register, supports wildcard.
        display : bool, optional
            disable print to screen. Default True
        """
        if display:
            print("Removing:`\n")
        match = self.list(names=names, display=display, relative=False)
        for k in match:
            _ = self.register.pop(k, None)
            _ = self.register_parent.pop(k, None)
            _ = self.register_indices.pop(k, None)
            _ = self.register_keys.pop(self.register_keys.index(k))

    def copy(self, names=None, shallow=False):
        """
        Make a copy (new TsDB instance) with the specified keys/names included.

        Parameters
        ----------
        names : str or list or tuple, optional
            Time series names filter that supports regular expressions.
        shallow: bool, optional
            If False (default), the copied TimeSeries objects do not point back to the TimeSeries instances in the
            source TsDB.

        Returns
        -------
        tsdb
            New TsDB instance.

        Notes
        -----
        If shallow is True, the TimeSeries objects in the new TsDB instance point back to the original objects. This
        implies that modifications to the copied TimeSeries are routed back to the source objects.
        The advantage of using ``shallow=True`` is that memory usage is limited. However, if in doubt use
        ``shallow=False`` (the default).

        Specified timeseries that are not preloaded (stored), will be loaded during this procedure.
        """
        new = TsDB(name=self.name)
        container = self.getm(names=names, store=True, fullkey=True)
        for key, ts in container.items():
            if shallow is False:
                ts = ts.copy()
            new.add(ts)
            new.register_parent[key] = self.register_parent[key]
            new.register_indices[key] = self.register_indices[key]
        return new

    def create_common_time(self, names=None, twin=None, maxdt=None, strict=False):
        """
        Creates common time array from: latest start time, earliest end time and minimum mean time step.

        Parameters
        ----------
        names : str or list or tuple, optional
            Time series names
        twin: tuple, optional
            If specified, the common time array within specified window (start, end) is created.
        maxdt: float, optional
            Max. time step desired. If minimum mean time step is larger than 'maxdt', the latter is used.
        strict: bool, optional
            If True, ValueError is raised if obtained time step deviates from specified/recommended time step.

        Returns
        -------
        array
            New time array.

        Notes
        -----
        Be careful when specifying the maximum time step.
        """
        container = self.getm(names=names, fullkey=True, store=False)
        timecheck = self._check_time_arrays(container)
        try:
            start, end, dt = timecheck["common"]
        except TypeError:
            raise ValueError("Could not create common time array "
                             "- check if earliest end time is before latest start time")
        if maxdt is not None:
            if dt > maxdt:
                dt = maxdt
        nt = int(round((end-start)/dt)) + 1
        common_time, _dt = np.linspace(start, end, nt, retstep=True)
        if strict is True and not round(dt - _dt, 4) == 0:
            raise ValueError("obtained 'dt' (%f) deviates from specified 'dt' (%s)" % (_dt, dt))
        if twin is not None:
            t_start, t_end = twin
            ind = (common_time >= t_start) & (common_time <= t_end)
            common_time = common_time[ind]
        return common_time

    def export(self, filename, names=None, delim="\t", skip_header=False, exist_ok=True, basename=True,
               verbose=False, **kwargs):
        """
        Export time series to file

        Parameters
        ----------
        filename : str
            File name including suffix.
        names : str or list or tuple, optional
            Time series names
        delim : str, optional
            column delimiter for column wise ascii file, default "\t"
        skip_header: bool, optional
            For ascii files, skip header with keys? Default is to include them on first line. This parameter is ignored
            for other file formats.
        exist_ok : bool, optional
            if false (the default), a FileExistsError is raised if target file already exists
        basename : bool, optional
            If true (the default), basename (no path/file info) will be exported to key file. If false, the name/path
            relative to common path is used. Also see notes below.
        verbose : bool, optional
            Print information
        kwargs : optional
            see documentation of :meth:`~qats.TimeSeries.get` method for available options

        Notes
        -----
        Currently implemented file formats
         - direct access format (.ts)
         - column-wise ascii file with single header line defining the keys(.dat) (comment character is #)

        The time series are resampled to a common time vector with a constant time step (sample rate). The minimum
        average time step of all the selected time series is applied. This is done before enforcing the specified
        time window.

        If `basename` is true, an exception is raised if two or more time series share the same basename. The solution
        is then to specify ``basename=False``, so that only the common part of the paths/identifiers of the specified
        time series is removed before writing the identifiers to key file (applies to .ts format) or ascii file header.
        Note that in this case the keys are modified so that path separators ('/' and '\\') are replaced by
        underscore ('_').
        """
        # todo: export(): include `header` parameter; text string that is included on top of key file

        # assert that filename is specified as string
        if not isinstance(filename, str):
            raise TypeError("Filename should be a str, not: %s" % type(filename))

        # assert that if the file exists the user has specifically allowed overwriting it
        if os.path.isfile(filename) and exist_ok is False:
            raise FileExistsError("The file '%s' already exists." % filename)

        # create non-existing directories
        dirname = self._path_dirname(filename)
        if dirname != "" and not os.path.exists(dirname):
            os.makedirs(dirname)

        # generate time series container
        '''
        Note:
        For most of the code below, it is convenient to use container generated by `getm()`. However; to evaluate
        whether time array is common (using `_check_time_arrays()`), we need TimeSeries objects in case `dtg_ref` 
        defined. The container is therefore generated as follows:
        1) Generate container of TimeSeries objects
        2) Modify container keys to export friendly keys
        3) Perform time array check (taking `dtg_ref` and  `**kwargs` into account)
        4) Convert container to container of arrays (same as output from `getm()`, taking **kwargs into account)
        '''
        # generate container, step 1 (generate container of TimeSeries objects)
        container = self.getm(names=names, fullkey=True, store=False)
        # generate container, step 2 (create export friendly keys)
        container = self._make_export_friendly_names(container, keep_basename=basename)
        # generate container, step 3 (perform time array check
        timecheck = self._check_time_arrays(container, **kwargs)
        # generate container, step 4 (convert to container of arrays)
        container = OrderedDict((k, v.get(**kwargs)) for k, v in container.items())

        # evaluate outcome of time array check, raise error if not common (time steps and time windows are not equal)
        if not timecheck["is_common"]:
            # create error message (principle: error+advice rather than automagic solution)
            # in other words; inform the user that the time arrays are not equal and suggest how to mitigate
            raise ValueError(timecheck["msg"])
        else:
            # all time arrays are equal, just use the time array from one of the TimeSeries objects
            common_time_array = container[list(container)[0]][0]

        # get file extension
        _, ext = os.path.splitext(filename)

        if ext == ".ts":    # write direct access file
            write_ts_data(filename, common_time_array, container)

        elif ext == ".dat":     # write ascii file
            write_dat_data(filename, common_time_array, container, delim=delim, skip_header=skip_header)

        elif ext == ".h5":
            write_sima_h5_data(filename, container)

        else:
            raise NotImplementedError("File format/type '%s' is not yet implemented." % ext)

        # print information
        if verbose:
            print(50 * "=")
            print("Exported %d records to file '%s'" % (len(container), filename))
            print(50 * "-")
            for key in container.keys():
                print(key)
            print(50 * "=")
        else:
            print("Exported %d records to file '%s'." % (len(container), filename))

    def get(self, name=None, ind=None, store=True):
        """
        Get (single) TimeSeries object.

        Parameters
        ----------
        name : str, optional
            Time series name
        ind : int, optional
            Time series index in database register.
        store : bool, optional
            Disable time series storage. Default is to store the time series objects first time it is read.

        Returns
        -------
        TimeSeries
            Time series

        Notes
        -----
        Either name or ind must be specified.

        Error is raised if zero or more than one match is obtained.

        Note that this method is somewhat similar to :meth:`~geta` but returns a TimeSeries object instead of arrays.
        Therefore this method does not support keyword arguments to be passed further to `TimeSeries.get()`.

        See also
        --------
        qats.TimeSeries
        geta, getda, getl, getm
        """
        # check that at least one of the required parameters is given,
        # and that non-compatible parameters are not combined
        if name is None and ind is None:
            raise TypeError("Either `name` or `ind` must be given")

        if name is not None and ind is not None:
            raise TypeError("Cannot combine parameters `ind` and `name`")

        # check type of specified parameters
        if name is not None and not isinstance(name, str):
            raise TypeError("Parameter `name` must be string if specified")

        if ind is not None and not isinstance(ind, int):
            raise TypeError("Parameter `ind` must be integer if specified")

        # return quickly if specified key (or index) matches a register entry and data is already loaded
        if name is not None:
            _ = self.list(names=name, display=False)
            if len(_) == 0:
                raise LookupError(f"No match found for '{name}'")

            elif len(_) > 1:
                raise ValueError(f"Name '{name}' is not unique. Got {len(_)} matches.")
            else:
                key = _[0]
        else:
            # ind is not None
            try:
                _ = self.register_keys[ind]
            except IndexError as err:
                raise IndexError(f"Index {ind} is not available (number of entries = {self.n})") from err
            else:
                key = _

        # check if data is read
        ts = self.register.get(key, None)
        if isinstance(ts, TimeSeries):
            return ts
        else:
            container = self.getm(names=name, ind=ind, fullkey=True, store=store)
            n = len(container)
            if n == 0:
                raise LookupError("No match found for specified name")
            elif n > 1:
                raise ValueError("More than one match found for specified name:"
                                 "\n    %s" % "\n    ".join(container.keys()))
            else:
                return container.popitem()[1]

    def geta(self, name=None, ind=None, store=True, **kwargs):
        """
        Get (single) time series as numpy arrays (tuple of time and data arrays).
        Optionally, the returned data array may be processed according to specified optional arguments (see `kwargs`).

        Parameters
        ----------
        name : str, optional
            Time series name
        ind : int, optional
            Time series index in database register.
        store : bool, optional
            Disable time series storage. Default is to store the time series objects first time it is read.
        kwargs : optional
            See documentation of :meth:`~qats.TimeSeries.get` method for available options

        Returns
        -------
        tuple
            Time and data arrays

        Notes
        -----
        Either name or ind must be specified.

        Error is raised if zero or more than one match is obtained.

        See also
        --------
        qats.TimeSeries, qats.TimeSeries.get
        get, getda, getl, getm
        """
        # use self.get() to avoid duplicate code
        try:
            ts = self.get(name=name, ind=ind, store=store)
        except TypeError:
            raise
        except LookupError:
            raise
        except ValueError:
            raise
        return ts.get(**kwargs)

    def getd(self, names=None, ind=None, store=True, fullkey=False):
        """
        Get dict of TimeSeries objects.

        Note that this method is identical to :meth:`~qats.TimeSeries.getm`.

        Parameters
        ----------
        names : str or list or tuple, optional
            Time series name(s), supports wildcard.
        ind : int or list, optional
            Index (or indices) of desired time series (index refers to index of key in list attribute `register_keys`).
            This parameter may not be combined with `keys` or `names`.
        store : bool, optional
            Disable time series storage. Default is to store the time series objects first time it is read.
        fullkey : bool, optional
            Use full key in returned container

        Returns
        -------
        dict
            TimeSeries objects

        See also
        --------
        qats.TimeSeries
        get, geta, getda, getl, getm
        """
        return self.getm(names=names, ind=ind, store=store, fullkey=fullkey)

    def getda(self, names=None, ind=None, store=True, fullkey=False, **kwargs):
        """
        Get dictionary of (numpy) arrays.
        Optionally, the returned data arrays may be processed according to specified optional arguments (see `kwargs`).

        Parameters
        ----------
        names : str or list or tuple, optional
            Time series names, supports wildcard.
        ind : int or list, optional
            Index (or indices) of desired time series (index refers to index of key in list attribute `register_keys`).
            This parameter may not be combined with `keys` or `names`.
        store : bool, optional
            Disable time series storage. Default is to store the time series objects first time it is read.
        fullkey : bool, optional
            Use full key in returned container
        kwargs : optional
            see documentation of :meth:`~qats.TimeSeries.get` method for available options

        Returns
        -------
        dict
            Each entry is a tuple with 2 arrays: time and data for each time series

        Notes
        -----
        When working on a large time series database it is recommended to set ``store=False`` to avoid too high memory
        usage. Then the TimeSeries objects will not be stored in the database, only their addresses.

        See also
        --------
        qats.TimeSeries, qats.TimeSeries.get
        get, geta, getl, getm
        """
        # read time series and put in ordered dictionary (reuse getm() to avoid duplicating code)
        container = OrderedDict((k, v.get(**kwargs)) for k, v in
                                self.getm(names=names, ind=ind, store=store, fullkey=fullkey).items())

        return container

    def getl(self, names=None, ind=None, store=True):
        """
        Get list of TimeSeries objects.

        Parameters
        ----------
        names : str or list or tuple, optional
            Time series names, supports wildcard.
        ind : int or list, optional
            Index (or indices) of desired time series (index refers to index of key in list attribute `register_keys`).
            This parameter may not be combined with `keys` or `names`.
        store : bool, optional
            Disable time series storage. Default is to store the time series objects first time it is read.
        Returns
        -------
        list
            TimeSeries objects

        Notes
        -----
        When working on a large time series database it is recommended to set ``store=False`` to avoid too high memory
        usage. Then the TimeSeries objects will not be stored in the database, only their addresses.

        See also
        --------
        qats.TimeSeries
        get, geta, getda, getm
        """
        # read time series and return in list (reuse getm() to avoid duplicating code)
        return list(self.getm(names=names, ind=ind, store=store).values())

    def getm(self, names=None, ind=None, store=True, fullkey=False):
        """
        Get (dictionary of) multiple TimeSeries objects.

        Parameters
        ----------
        names : str or list or tuple, optional
            Time series name(s), supports wildcard.
        ind : int or list, optional
            Index (or indices) of desired time series (index refers to index of key in list attribute `register_keys`).
            This parameter may not be combined with `keys` or `names`.
        store : bool, optional
            Disable time series storage. Default is to store the time series objects first time it is read.
        fullkey : bool, optional
            Use full key in returned container

        Returns
        -------
        dict
            TimeSeries objects

        Notes
        -----
        When working on a large time series database it is recommended to set ``store=False`` to avoid too high memory
        usage. Then the TimeSeries objects will not be stored in the database, only their addresses.

        Note that this method is somewhat similar to :meth:`~getda` but returns TimeSeries objects instead of numpy
        arrays. Therefore this method does not support keyword arguments to be passed further to TimeSeries.get().

        See also
        --------
        qats.TimeSeries
        get, geta, getda, getl
        """
        # check that non-compatible parameters are not combined
        if ind is not None and names is not None:
            raise TypeError("Cannot combine parameters `ind` and `names`")

        # get absolute keys
        if ind is None:
            keys = self.list(names=names, display=False, relative=False)
        else:
            # generate keys directly from indices
            if isinstance(ind, int):
                ind = [ind]
            try:
                keys = [self.register_keys[i] for i in ind]
            except IndexError as err:
                raise IndexError("one or more index is out of range (number of entries = %d)" % self.n) from err
            except TypeError as err:
                raise TypeError("Parameter `ind` must be integer or list of integers") from err

        if fullkey:
            # use full key in returned container
            retkeys = keys
        else:
            # todo: consider including the '\\' in self.common (currently I am not sure what is the best)
            # create shorter keys (relative to db common path)
            _common = self.common + os.path.sep
            retkeys = [k.replace(_common, "") for k in keys]

        # read time series and put in ordered dictionary
        container = self._read(keys, retkeys=retkeys, store=store)

        return container

    def is_common_time(self, names=None, twin=None):
        """
        Check if time array is common.

        Parameters
        ----------
        names : str or list or tuple, optional
            Time series names
        twin: tuple, optional
            Time window (start, end) to consider.

        Returns
        -------
        bool
            True if common time array (within specified time window), otherwise False.
        """
        container = self.getm(names=names, fullkey=True, store=False)
        timecheck = self._check_time_arrays(container, twin=twin)
        return timecheck["is_common"]

    def list(self, names=None, display=False, relative=False):
        """
        List time series in database by id/key

        Parameters
        ----------
        names : str or list or tuple, optional
            Time series names filter that supports regular expressions, default all time series will be listed
        display : bool, optional
            Disable print to screen, default False
        relative : bool, optional
            Truncate time series names to unique part i.e. path relative to common path of all time series.

        Returns
        -------
        list
            Time series names

        Notes
        -----
        Full identifier/key is obtained by joining the common path of all time series in db and the unique part of the
        identifiers.
        """
        def _remove_special_characters(strings):
            """
            Remove characters with special meaning in regular expressions

            Parameters
            ----------
            strings : tuple or list
                Strings from which special characters are to be removed

            Returns
            -------
            list :
                Strings with special characters removed

            """
            out = []
            for s in copy.copy(strings):
                # handle regex special characters
                s = s.replace("[", ":[:")  # to avoid troubles when "[" is inserted to handle "]"
                s = s.replace("]", "[]]")
                s = s.replace(":[:", "[[]")
                s = s.replace("^", "[^]")
                s = s.replace("(", "[(]")
                s = s.replace(")", "[)]")
                out.append(s)
            return out

        # full names in db register
        keys_in_register = copy.copy(self.register_keys)

        # get common path
        common = self.common

        # handle types and add leading wildcard to enable pattern matching across paths
        # Note: wildcard should not be prepended:
        #           - if there is no common path
        #           - for names starting with the common path (name is then a full path)
        if names is not None:
            if common == "":
                # no common path -> prefix should not be included (and is not needed)
                _prefix = ""
            else:
                # add prefix *\\ (or */ in unix)
                _prefix = '*' + os.path.sep
            if isinstance(names, str):
                if not names.startswith(common):
                    names = _prefix + names
                names = [names]
            elif type(names) in (list, tuple):
                names = [_prefix + n if not n.startswith(common) else n for n in names]
            else:
                raise TypeError(f"Parameter `names` should be of type str/list/tuple, not {type(names)}")

        # generate list of names/keys
        if names is None:
            # all time series names if no patterns are specified
            match = keys_in_register
        else:
            # match time series keys with specified time series name patterns
            match = []
            for name in _remove_special_characters(names):
                match.extend(fnmatch.filter(keys_in_register, name))

        if relative:
            match = [self._path_relpath(_, common) for _ in match]

        if display:
            print()
            print("=============================================================================================")
            if relative:
                print(f"Common path : {common}")
                print("---------------------------------------------------------------------------------------------")
            if len(match) > 0:
                print("\n".join(match))
            else:
                print("(no match found for specified names/names)")
            print("=============================================================================================")

        return match

    def load(self, filenames, read=False, verbose=False):
        """
        Load time series from files

        Parameters
        ----------
        filenames : str or list or tuple
            File names including suffix. Wildcards can also be used.
        read: bool, optional
            If True, all time series are read from file and stored. The default is that they are read from file
            when requested by any of the `get` methods.
        verbose : bool, optional
            If True, print information to screen.

        Notes
        -----
        `read=True` may be time consuming and require high memory usage if applied for large files with many
        time series. However, if you will work with all the time series on files of moderate size, `read=True` can
        provide efficiency as you only access the file(s) once.

        If time series names contain path separators ('/' or '\\'), these must be enclosed within square brackets (e.g.
        as the slash in "response[m/s]").
        """
        if isinstance(filenames, list) or isinstance(filenames, tuple):
            # expect that iterable contains set of file names, stored with absolute path
            files = [os.path.abspath(f) for f in filenames]
        elif isinstance(filenames, str):
            # string is interpreted as a filename, possibly with wildcards, stored with absolute path
            files = [os.path.abspath(f) for f in glob.glob(filenames)]
        else:
            raise TypeError("files should be either str/tuple/list, not: %s" % type(filenames))

        # If the specified filenames does not exist glob.glob returns an empty list
        if len(files) < 1:
            raise FileExistsError("Path does not exist: %s" % filenames)

        # read time series names and possibly also the data
        for thefile in files:
            fext = os.path.splitext(thefile)[-1]
            basename = os.path.basename(thefile)
            dirname = os.path.dirname(thefile)

            if not os.path.isfile(thefile):
                raise FileExistsError("Object is not a file: %s" % thefile)

            if fext == '.ts':
                # direct access format without info array
                names = read_ts_names(thefile.replace(fext, '.key'))

            elif fext == '.tda':
                # simo s2x direct access format (with info array)
                names = read_tda_names(thefile.replace(fext, '.txt'))

            elif fext == '.asc':
                # simo-riflex, sima ascii
                names = read_sima_names(os.path.join(dirname, 'key_' + basename.replace(fext, '.txt')))

            elif fext == '.bin':
                _ = os.path.join(dirname, 'key_' + basename.replace(fext, '.txt'))
                if thefile.endswith('witurb.bin') or thefile.endswith('blresp.bin'):
                    # wind turbine data and blade response is stored on .bin files but the corresponding
                    # key file has a different structure than the ones associated with 'elmfor' and 'noddis'
                    # .bin files
                    names = read_sima_wind_names(_)
                else:
                    # riflex/simo, sima direct access format
                    names = read_sima_names(_)

            elif fext == '.dat':
                # plain column wise ascii format
                names = read_dat_names(thefile)

            elif fext == '.mat':
                # SINTEF Ocean test data export format based on Matlab .mat files.
                _tk, names = read_mat_names(thefile)
                self._timekeys[thefile] = _tk   # remember the name of the time array

            elif fext in ('.h5', '.hdf5'):
                # sima h5
                names = read_sima_h5_names(thefile)

            elif fext == '.csv':
                # column wise csv
                names = read_csv_names(thefile)

            elif fext == '.tdms':
                # National Instrument structured binary file format
                names = read_tdms_names(thefile)

            else:
                raise NotImplementedError("Invalid file type: %s" % thefile)

            # update database register: unique id (key) consists of filename and time series name
            # before the time series is loaded the register store the time series index on the file
            # j +1 since record 0 is the time vector
            for j, name in enumerate(names):
                key = os.path.join(thefile, name)
                # None until the time series is read and stored
                self.register[key] = None
                # Parent, i.e. source file
                self.register_parent[key] = thefile
                # Time series index on file, to speed up reading the time series, dummy for .mat files, +1 to skip time
                ind = j + 1 if fext not in ('.h5', '.hdf5', '.mat') else None
                self.register_indices[key] = ind
                # time series names in the order the associated time series where loaded
                self.register_keys.append(key)

            if read is True:
                keys = [os.path.join(thefile, name) for name in names]
                _ = self._read(keys, store=True)

            if verbose:
                print("Loaded %d records from file '%s'." % (len(names), thefile))
                print('\n'.join(names))

    def plot(self, names=None, figurename=None, show=None, num=1, store=True, **kwargs):
        """
        Plot time series traces.

        Parameters
        ----------
        names : str/list/tuple, optional
            Time series names
        figurename : str, optional
            Save figure to file 'figurename' instead of displaying on screen.
        show : bool, optional
            Show figure? Defaults to False if `figurename` is specified, otherwise True.
        num : int, optional
            Matplotlib figure number. Defaults to 1.
        store : bool, optional
            Disable time series storage. Default is to store the time series objects first time it is read.
        kwargs : optional
            See documentation of TimeSeries.get() method for available options

        Notes
        -----
        When working on a large time series database it is recommended to set ``store=False`` to avoid too high memory
        usage. Then the TimeSeries objects will not be stored in the database, only their addresses.

        See also
        --------
        qats.TimeSeries

        """
        # todo: add possibility for subplots in plot method. nsub=None (int), sharex=True.
        # todo: consider need for `keep_order` parameter when plotting
        # dict with numpy arrays: time and data
        container = self.getda(names=names, store=store, **kwargs)

        plt.figure(num=num)
        for k, v in container.items():
            label = k  # todo: more readable label, e.g. remove commonpath
            plt.plot(v[0], v[1], label=label)

        plt.xlabel('Time (s)')
        plt.grid()
        plt.legend()
        if figurename is not None:
            plt.savefig(figurename)
        if show is True or (show is None and figurename is None):
            plt.show()

    def plot_psd(self, names=None, figurename=None, show=None, num=1, store=True, **kwargs):
        """
        Plot time series power spectral density.

        Parameters
        ----------
        names : str/list/tuple, optional
            Time series names
        figurename : str, optional
            Save figure to file 'figurename' instead of displaying on screen.
        show : bool, optional
            Show figure? Defaults to False if `figurename` is specified, otherwise True.
        num : int, optional
            Matplotlib figure number. Defaults to 1.
        store : bool, optional
            Disable time series storage. Default is to store the time series objects first time it is read.
        kwargs : optional
            see documentation of TimeSeries.get() method for available options

        Notes
        -----
        When working on a large time series database it is recommended to set ``store=False`` to avoid too high memory
        usage. Then the TimeSeries objects will not be stored in the database, only their addresses.

        See also
        --------
        qats.TimeSeries

        """
        # dict with TimeSeries objects
        container = self.getm(names=names, store=store)

        plt.figure(num=num)
        for k, v in container.items():
            f, p = v.psd(**kwargs)
            plt.plot(f, p, label=k)

        plt.xlabel('Frequency (Hz)')
        plt.ylabel('Power spectral density')
        plt.grid()
        plt.legend()
        if figurename is not None:
            plt.savefig(figurename)
        if show is True or (show is None and figurename is None):
            plt.show()

    def plot_cycle_range(self, names=None, n=200, w=None, bw=1., figurename=None, show=None, num=1, store=True,
                         **kwargs):
        """
        Plot cycle range versus number of occurrences.

        Parameters
        ----------
        names : str/list/tuple, optional
            Time series names
        n : int, optional
            Group by cycle range in *n* equidistant bins.
        w : float, optional
            Group by cycle range in *w* wide equidistant bins. Overrides *n*.
        bw : float, optional
            Bar width, expressed as ratio of bin width.
        figurename : str, optional
            Save figure to file 'figurename' instead of displaying on screen.
        show : bool, optional
            Show figure? Defaults to False if `figurename` is specified, otherwise True.
        num : int, optional
            Matplotlib figure number. Defaults to 1.
        store : bool, optional
            Disable time series storage. Default is to store the time series objects first time it is read.
        kwargs : optional
            see documentation of TimeSeries.get() method for available options

        Notes
        -----
        When working on a large time series database it is recommended to set store=False to avoid too high memory
        usage. Then the TimeSeries objects will not be stored in the database, only their addresses.

        See Also
        --------
        TsDB.rfc, TsDB.plot_cycle_rangemean, TimeSeries.rfc, rainflow.count_cycles, rainflow.rebin_cycles

        """
        # dict with TimeSeries objects
        container = self.getm(names=names, store=store)

        # ensure that rebinning will be done
        assert (n is not None) or (w is not None), "Cycles must be rebinned for this plot - either 'n' or 'w' must " \
                                                   "be different from None"

        plt.figure(num=num)
        for k, v in container.items():
            # extract cycles
            cycles = v.rfc(**kwargs)

            # rebin cycles
            cycles = rebin_cycles(cycles, binby='range', n=n, w=w)

            r, _, c = zip(*cycles)  # unpack range and count pairs, ignore mean value
            dr = r[1] - r[0]        # bin width, used as basis for bar width
            plt.bar(r, c, dr * bw, label=k, alpha=0.4)

        plt.xlabel('Cycle range')
        plt.ylabel('Cycle count (-)')
        plt.grid()
        plt.legend()
        if figurename is not None:
            plt.savefig(figurename)
        if show is True or (show is None and figurename is None):
            plt.show()

    def plot_cycle_rangemean(self, names=None, n=None, w=None, figurename=None, show=True, num=1, store=True, **kwargs):
        """
        Plot cycle range-mean versus number of occurrences.

        Parameters
        ----------
        names : str/list/tuple, optional
            Time series names
        n : int, optional
            Group by cycle range in *n* equidistant bins.
        w : float, optional
            Group by cycle range in *w* wide equidistant bins. Overrides *n*.
        figurename : str, optional
            Save figure to file 'figurename' instead of displaying on screen.
        show : bool, optional
            Show figure? Defaults to False if `figurename` is specified, otherwise True.
        num : int, optional
            Matplotlib figure number. Defaults to 1.
        store : bool, optional
            Disable time series storage. Default is to store the time series objects first time it is read.
        kwargs : optional
            see documentation of TimeSeries.get() method for available options

        Notes
        -----
        When working on a large time series database it is recommended to set ``store=False`` to avoid too high memory
        Cycle means are represented by weighted averages in each bin.

        When working on a large time series database it is recommended to set store=False to avoid too high memory
        usage. Then the TimeSeries objects will not be stored in the database, only their addresses.

        See also
        --------
        qats.TimeSeries

        """
        # dict with TimeSeries objects
        container = self.getm(names=names, store=store)

        plt.figure(num=num)
        for k, v in container.items():
            # extract cycles
            cycles = v.rfc(**kwargs)

            # rebin cycles
            if (n is not None) or (w is not None):
                cycles = rebin_cycles(cycles, binby='range', n=n, w=w)

            ranges, means, counts = zip(*cycles)      # unpack range and count pairs, ignore mean value
            plt.scatter(means, ranges, s=[2. * c for c in counts], label=k, alpha=0.4)  # double marker size for improved readability

        plt.xlabel('Cycle mean')
        plt.ylabel('Cycle range')
        plt.grid()
        plt.legend()
        if figurename is not None:
            plt.savefig(figurename)
        if show is True or (show is None and figurename is None):
            plt.show()

    def rename(self, name, newname):
        """
        Rename a timeseries (and update register accordingly).

        Parameters
        ----------
        name : str
            Unique time series name (pattern)
        newname : str
            New time series name (i.e. unique id/key less the path of parent file).

        Notes
        -----
        A single match must be obtained for the specified `name`.
        """
        # find matching time series names
        names = self.list(names=name, display=False, relative=False)
        n = len(names)
        if n == 0:
            raise LookupError("No match found for specified name")
        elif n > 1:
            raise ValueError("More than one match found for specified name:"
                             "\n    %s" % "\n    ".join(names))

        # define new key, and check that it doesn't exist
        # todo: consider replace (parent --> "") instead of dirname (only relevant for ts from .h5?)
        oldkey = names[0]
        newkey = os.path.join(self._path_dirname(oldkey), newname)
        if newkey in self.register_keys:
            raise ValueError(f"New key already exists: {newkey}")

        # rename (change register)
        self.register[newkey] = self.register.pop(oldkey)
        self.register_parent[newkey] = self.register_parent.pop(oldkey)
        self.register_indices[newkey] = self.register_indices.pop(oldkey)
        self.register_keys[self.register_keys.index(oldkey)] = newkey

        # rename TimeSeries instance if it is pre-loaded
        ts = self.register.get(newkey, None)
        if isinstance(ts, TimeSeries):
            ts.name = newname

        return

    def stats(self, statsdur=10800., names=None, ind=None, store=True, fullkey=False, **kwargs):
        """
        Get statistics for time series processed according to parameters

        Parameters
        ----------
        statsdur : float, optional
            Duration in seconds for estimation of extreme value distribution (Gumbel) from peak distribution (Weibull).
            Default is 3 hours.
        names : str or list or tuple, optional
            Time series names
        ind : int or list, optional
            Time series indices in database
        store : bool, optional
            Disable time series storage. Default is to store the time series objects first time it is read.
        fullkey : bool, optional
            Use full key in returned container
        kwargs : optional
            see documentation of TimeSeries.stats() and TimeSeries.get() methods for available options

        Returns
        -------
        dictionary
            Each entry is a dictionary with statistics

        Notes
        -----
        Full unique path/key is obtained by joining the common part of the paths/keys and the unique part of the keys

        When working on a large time series database it is recommended to set ``store=False`` to avoid too high memory
        usage. Then the TimeSeries objects will not be stored in the database, only their addresses.

        See also
        --------
        qats.TimeSeries.stats, qats.TimeSeries.get

        Examples
        --------
        To get sample statistics and 3-hour extreme (default) value statistics for time series in database

        >>> from qats import TsDB
        >>> db = TsDB.fromfile('filename')
        >>> stats = db.stats()

        To get sample statistics and 1-hour (3600 seconds) extreme value statistics for time series in database which
        names starts with 'Mooring'.

        >>> stats = db.stats(statsdur=3600., names="Mooring*")

        To get sample statistics and 3-hour extreme value statistics for time series in database which names starts with
        'Mooring' low-pass filtered at 0.025 Hz.

        >>> stats = db.stats(names="Mooring*", filterargs=("lp", 0.025))

        To ignore the transient part of the time series, time window may be specified:

        >>> stats = db.stats(names="Mooring*", twin=(1000., 1e12), filterargs=("lp", 0.025))

        """
        # todo: create entry in dictionary with meta data such as applied time window, filter etc.
        # read time series and put in ordered dictionary (reuse getm() to avoid duplicating code)
        container = OrderedDict((k, v.stats(statsdur=statsdur, **kwargs)) for k, v in
                                self.getm(names=names, ind=ind, store=store, fullkey=fullkey).items())

        return container

    def stats_dataframe(self, statsdur=10800., names=None, ind=None, store=True, fullkey=False, **kwargs):
        """
        Get Pandas Dataframe with time series statistics.

        Parameters
        ----------
        statsdur : float, optional
            Duration in seconds for estimation of extreme value distribution (Gumbel) from peak distribution (Weibull).
            Default is 3 hours.
        names : str or list or tuple, optional
            Time series names
        ind : int or list, optional
            Time series indices in database
        store : bool, optional
            Disable time series storage. Default is to store the time series objects first time it is read.
        fullkey : bool, optional
            Use full key in returned container
        kwargs : optional
            see documentation of TimeSeries.stats() and TimeSeries.get() methods for available options

        Returns
        -------
        Dataframe
            Statistics for each time series organized in columns.

        See also
        --------
        qats.TsDB.stats, pandas.Dataframe

        Examples
        --------
        >>> from qats import TsDB
        >>> db = TsDB.fromfile('filename')
        >>> df = db.stats_dataframe()
        >>> # print first part of dataframe
        >>> print(df.head())

        >>> # transpose dataframe to get statistics for each time series in rows instead of columns
        >>> df = df.transpose(copy=True)
        """
        df = pd.DataFrame(self.stats(statsdur=statsdur, names=names, ind=ind, store=store, fullkey=fullkey, **kwargs))

        return df

    def update(self, tsdb, names=None, shallow=False):
        """
        Update TsDB with speicified keys/names from other TsDB instance.

        Parameters
        ----------
        tsdb: TsDB
            TsDB instance to update from.
        names : str or list or tuple, optional
            Time series names
        shallow: bool, optional
            If False (default), the copied TimeSeries objects do not point back to the TimeSeries instances in the
            source TsDB.

        """
        if not isinstance(tsdb, TsDB):
            raise TypeError("expected TsDB instance, got: %s" % type(tsdb))

        container = tsdb.getm(names=names, fullkey=True)
        for key, ts in container.items():
            if key in self.register.keys():
                raise KeyError("The specified key is not unique: %s" % key)

            if shallow is False:
                ts = ts.copy()

            self.register[key] = ts
            self.register_keys.append(key)
            self.register_parent[key] = tsdb.register_parent[key]
            self.register_indices[key] = tsdb.register_indices[key]

        return



"""
Readers for various direct access formatted time series files
"""
import os
from array import array
from struct import pack, unpack

import numpy as np


def read_ts_names(path):
    """
    Read time series names from key-file associated with the direct access (binary) time series file format '.ts'.

    Parameters
    ----------
    path : str
        Key file path

    Returns
    -------
    list
        Time series names

    Notes
    -----
    Names are stored on ASCII file as one name per line. The file is terminated by END.

    The associated time series are stored on a binary direct access file with suffix '.ts'.
    """
    # read line-separated keys from ascii file
    names = _read_names(path)

    # skip the key referring to the time array (always the first one on the ascii file)
    return names[1:]


def read_tda_names(path):
    """
    Read time series names from key-file associated with the direct access (binary) time series file format '.tda'.

    Parameters
    ----------
    path : str
        Key file path

    Returns
    -------
    list
        Time series names

    Notes
    -----
    The associated time series are stored on a binary direct access file with suffix '.tda'.
    """
    # Read line-separated names from ascii file
    names = _read_names(path)

    # Skip the keys referring to the
    # - info array (always the first one on the ascii file)
    # - time array (always the second one on the ascii file)
    return names[2:]


def read_dis_names(path):
    """
    Read time series names from key-file associated with the direct access (binary) cycle distribution file format '.dis'.

    Parameters
    ----------
    path : str
        Key file path

    Returns
    -------
    list
        Time series names

    Notes
    -----
    The associated cycle distributions are stored on a binary direct access file with suffix '.dis'.
    """
    # read line-separated names from ascii file
    names = _read_names(path)

    return names


def _read_names(path):
    """
    Read line-separated time series names from ascii file

    Parameters
    ----------
    path : str
        Key file path

    Returns
    -------
    list
        Time series names

    Notes
    -----
    Keys are stored on ASCII file as one key per line. The file is terminated by END.
    """
    with open(os.path.join(path, path), 'r') as f:
        names = [l.strip() for l in f if not l.startswith(("**", "'")) and not l.upper().strip() == "END"]

    return names


def read_ts_data(path, ind=None, verbose=False):
    """
    Read time series from binary direct access formatted '.ts' file.

    Parameters
    ----------
    path : str
        File path (relative or absolute)
    ind : list of integers, optional
        Index of the requested records on the file. By default all time series are read.
    verbose : bool, optional
        Increase verbosity

    Returns
    -------
    array
        Time and data

    Examples
    --------
    >>> data = read_ts_data('data.ts')
    >>> t = data[0,:]   # time
    >>> x1 = data[1,:]  # first data series

    Notes
    -----
    Note that `0` is the index of the time array. If ``ind`` is specified, time array is only included if
    `0` is included in the specified indices.
    """
    # the format specifier for decoding binary integers is 'i'
    data = _read_data(path, ifmt='i', ind=ind, verbose=verbose)

    return data


def read_tda_data(path, ind=None, verbose=False):
    """
    Read time series from binary direct access formatted '.tda' file.

    Parameters
    ----------
    path : str
        File path (relative or absolute)
    ind : list of integers, optional
        Index of the requested records on the file. By default all time series are read.
    verbose : bool, optional
        Increase verbosity

    Returns
    -------
    array
        Time and data

    Examples
    --------
    >>> data = read_tda_data('data.tda')
    >>> t = data[0,:]   # time
    >>> x1 = data[1,:]  # first data series

    Notes
    -----
    Note that `0` is the index of the time array. If ``ind`` is specified, time array is only included if
    `0` is included in the specified indices.
    """
    # the format specifier for decoding binary integers is 'i'
    data = _read_data(path, ifmt='f', ind=ind, verbose=verbose)

    return data


def read_dis_data(path, ind=None, verbose=False):
    """
    Read cycle distributions from binary direct access formatted '.dis' file.

    Parameters
    ----------
    path : str
        File path (relative or absolute)
    ind : list of integers, optional
        Index of the requested records on the file. By default all time series are read.
    verbose : bool, optional
        Increase verbosity

    Returns
    -------
    array
        Cycle widths and counts

    Notes
    -----
    Note that `0` is the index of the time array. If ``ind`` is specified, time array is only included if
    `0` is included in the specified indices.
    """
    # the format specifier for decoding binary integers is 'i'
    data = _read_data(path, ifmt='i', ind=ind, verbose=verbose)

    return data


def _read_data(path, ifmt, ind=None, verbose=False):
    """
    Read direct access formatted files with time series or cycle distributions

    Parameters
    ----------
    path : str
        File path (relative or absolute)
    ifmt : str
        Format to decode binary integers
    ind : list of integers, optional
        Index of the requested records on the file. Note that `0` is the index of the time array. All data
        is read by default.
    verbose : bool, optional
        Increase verbosity

    Returns
    -------
    array
        Data

    Notes
    -----
    If ``ind`` is specified, time array is only included if `0` is included in the specified indices.

    If ``ind`` is specified, the response arrays (1-D) are stored in the returned 2-D array in the same order as
    specified. I.e. if ``ind=[0,10,2,3]``, then the response array with index `10` on .ts file is obtained from
    ``arr[1,:]``.

    For reading of .dis files, keep in mind that there is no common time array or similar for each data set. For
    example; if one intends to request data set number three, one would need to request indices 4 and 5 (first set is
    stored on indices 0 and 1, second on 2 and 3, etc.)

    The correct integer format varies with OS and the software exporting the data to file.
        - .ts on Windows  : 'i', 'l'
        - .ts on Linux    : 'i'
        - .tda on Windows : 'f'
        - .tda on Linux   : ?
        - .dis on Windows : 'i'
        - .dis on Linux : ?
    """
    if verbose:
        print('Reading %s ...' % path)

    # indices --> list or None
    if isinstance(ind, int):
        ind = [ind]

    # check extension
    if ifmt not in ("i", "l", "f"):
        raise ValueError(f"Invalid format for decoding binary integers: {ifmt}")

    try:
        with open(path, 'rb') as f:
            # read ts-file information (1st line)
            nbytes = 4
            s = f.read(nbytes)
            ndat = unpack(ifmt, s)[0]  # number of time steps per array

            # the code below works generally, but for some direct access files: nts=nrec-1
            # therefore, `nrec` is based on parsing the file instead
            #
            #   s = f.read(nbytes)
            #   nrec = unpack(ifmt, s)[0]  # number of records(including info record and time record)
            #   nts = nrec-2

            f.seek(0, 2)  # go to end of file
            epos = f.tell()  # get end position (in bytes)
            nrec = epos / (ndat * nbytes)  # no. of "rows"
            nts = nrec - 2  # excl. time array
            # make integers (necessary for ifmt='f')
            ndat = int(ndat)
            nts = int(nts)

            # extract all keys/indices, if <ind> not specified
            if ind is None:
                ind = list(range(nts + 1))  # (including time array)
            else:
                # check number of keys (+1 is due to bug in some direct access files), if specified by user
                assert max(ind) <= nts, "Requested time series no. %d, but there are only %d time series on file" % \
                                        (max(ind), nts)

            # define position
            pos = dict(zip(ind, range(len(ind))))

            # info
            if verbose:
                print('------------------------------------')
                print('nrec (no. of keys on .ts)   : %d' % nts)
                print('ndat (no. of time steps)    : %d' % ndat)
                print('number of keys to read      : %d  (%s)' % (len(ind), 'including time array index 0 is specified'))
                # minus 1 due to time array #len(keys)

            # initiate array
            arr = np.zeros((len(ind), ndat))
            # format string
            fmtstr = 'f' * ndat
            # go to second "row"
            f.seek(nbytes * ndat)
            # read arrays
            j = 0
            for i in range(nts + 1):
                if i in ind:
                    s = f.read(nbytes * ndat)
                    p = pos[i]
                    arr[p, :] = np.array(unpack(fmtstr, s))
                    j += 1
                else:
                    # seek position of next row
                    f.seek(4 * ndat, 1)  # (1 ==> seek relative to current position)

    except OverflowError:
        raise
    except AssertionError:
        raise

    return arr


def write_ts_data(path, time: np.ndarray, data: dict):
    """
    Write time series direct access formatted file.

    Parameters
    ----------
    path : str
        File path
    time : array
        Time
    data : dict
        Time series data name vs. values
    """
    # deduct name of key file
    base, _ = os.path.splitext(path)
    keyfilename = base + ".key"

    # open key file (ascii) and .ts file (binary)
    # todo: check if 'w' or 'wb' should be used, seems like fts.write(array(....)) expects a bytearray
    fts = open(path, "wb")
    fkey = open(keyfilename, "w")
    try:
        # number of data points in a single time series
        ndat = time.size

        # number of records (number of time series + header + time vector)
        nrec = len(data) + 2

        # write meta info to first record
        fts.write(pack("ii", ndat, nrec))

        # zero pad the first record
        fts.write(array("i", [0] * (ndat - 2)))

        # time array
        fkey.write("time\n")
        fts.write(array("f", time))

        # time series
        for key, arr in data.items():
            # write key to key file
            fkey.write("%s\n" % key)

            # write time series array to ts file (position 1 refers to time series data)
            fts.write(array("f", arr[1]))

    except Exception:
        raise RuntimeError("Exception encountered when writing data to file '%s'." % filename)

    finally:
        # end key file and close file pointers
        fkey.write("END\n")
        fkey.close()
        fts.close()





#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module with custom widgets

@author: perl
"""
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QTableWidget, QTableWidgetItem, QTabWidget


class CustomTabWidget (QTabWidget):
    """
    Custom tab widget to enable closing tabs
    """
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTabsClosable(True)
        self.tabCloseRequested.connect(self.close_tab)

    def close_tab(self, index):
        """
        Close tab by index
        """
        current_widget = self.widget(index)
        current_widget.deleteLater()
        self.removeTab(index)


class CustomTableWidget(QTableWidget):
    """ Placeholder for possible future customization of QTableWidget.
    Currently behaves exactly as its parent. """


class CustomTableWidgetItem(QTableWidgetItem):
    """ To enable table sorting based on value (not string) """
    # based on: https://stackoverflow.com/questions/11938459/sorting-in-pyqt-tablewidget

    def __lt__(self, other):
        if isinstance(other, QTableWidgetItem):
            left_var = self.data(Qt.EditRole)
            right_var = other.data(Qt.EditRole)
            try:
                return float(left_var) < float(right_var)
            except (ValueError, TypeError):
                return left_var < right_var

        return super(CustomTableWidgetItem, self).__lt__(other)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:class:`GumbelMin` class and functions related to Gumbel (minima) distribution.
"""
import numpy as np
from matplotlib.pyplot import (figure, grid, legend, plot, savefig, show,
                               xlabel, ylabel, ylim, yticks)
from scipy.optimize import fsolve, leastsq
from scipy.special import zetac

from .empirical import empirical_cdf
from .gumbel import _euler_masceroni as em

# todo: move fit methods e.g. _msm from class to standalone functions (importable)
# todo: check fit procedures (read up once more and check implementation)
# todo: create unit tests


class GumbelMin(object):
    """
    The Gumbel minima distribution.

    The cumulative distribution function is defined as::

        F(x) = 1 - exp{-exp[(x-a)/b]}

    where `a` is location parameter and `b` is the scale parameter.

    Parameters
    ----------
    loc : float
        Gumbel location parameter.
    scale : float
        Gumbel scale parameter.
    data : array_like, optional
        Sample data, used to establish empirical cdf and is included in plots.
        To fit the Gumbel distribution to the sample data, use :meth:`GumbelMin.fit`.

    Attributes
    ----------
    loc : float
        Gumbel location parameter.
    scale : float
        Gumbel scale parameter.
    data : array_like
        Sample data.

    Examples
    --------
    To initiate an instance based on parameters, use:

    >>> from qats.stats.gumbelmin import GumbelMin
    >>> gumb = GumbelMin(loc, scale)

    If you need to establish a Gumbel instance based on a sample data set, use:

    >>> gumb = GumbelMin.fit(data, method='msm')

    References
    ----------
    1. Bury, K.V. (1975) Statistical models in applied science. Wiley, New York

    2. Haver, S. (2007), "Bruk av asymptotiske ekstremverdifordelinger"

    3. `Plotting positions <http://en.wikipedia.org/wiki/Q%E2%80%93Q_plot>`_, About plotting positions

    4. `Usable estimators for parameters in Gumbel distribution
       <http://stats.stackexchange.com/questions/71197/usable-estimators-for-parameters-in-gumbel-distribution>`_

    5. `Bootstrapping statistics <https://en.wikipedia.org/wiki/Bootstrapping_(statistics)>`_

    """

    def __init__(self, loc=None, scale=None, data=None):
        self.location = loc
        self.scale = scale

        if data is not None:
            self.data = np.array(data)
        else:
            self.data = None

    @property
    def cov(self):
        """
        Distribution coefficient of variation (C.O.V.)

        Returns
        -------
        c : float
            distribution c.o.v.
        """
        return self.std / self.mean

    @property
    def ecdf(self):
        """
        Median rank empirical cumulative distribution function associated with the sample

        Notes
        -----
        Gumbel recommended the following mean rank quantile formulation Pi = i/(n+1).
        This formulation produces a symmetrical CDF in the sense that the
        same plotting positions will result from the data regardless of
        whether they are assembled in ascending or descending order.

        A more sophisticated median rank formulation Pi = (i-0.3)/(n+0.4) approximates the
        median of the distribution free estimate of the sample variate to about
        0.1% and, even for small values of n, produces parameter estimations
        comparable to the result obtained by maximum likelihood estimations (Bury, 1999, p43)
        A median rank method, pi=(i-0.3)/(n+0.4), is chosen to approximate the mean of the distribution [2]

        The empirical cdf is also used as plotting positions when plotting the sample
        on probability paper.
        """
        try:
            p = empirical_cdf(self.data.size, kind='median')
            return p
        except TypeError:
            print("The sample is not defined.")

    @property
    def kurt(self):
        """
        Distribution kurtosis

        Returns
        -------
        k : float
            distribution kurtosis
        """
        try:
            k = 12. / 5.
            return k
        except TypeError:
            print("Distribution parameters are not defined.")

    @property
    def mean(self):
        """
        Distribution mean value

        Returns
        -------
        m : float
            distribution mean value
        """
        try:
            m = self.location - self.scale * em()
            return m
        except TypeError:
            print("Distribution parameters are not defined.")

    @property
    def median(self):
        """
        Distribution median value

        Returns
        -------
        m : float
            distribution median value
        """
        try:
            m = self.location + self.scale * np.log(np.log(2.))
            return m
        except TypeError:
            print("Distribution parameters are not defined.")

    @property
    def mode(self):
        """
        Distribution mode value

        Returns
        -------
        m : float
            distribution mode value
        """
        try:
            m = self.location
            return m
        except TypeError:
            print("Distribution parameters are not defined.")

    @property
    def std(self):
        """
        Distribution standard deviation

        Returns
        -------
        s : float
            distribution standard deviation
        """
        try:
            s = np.pi * self.scale / np.sqrt(6)
            return s
        except TypeError:
            print("Distribution parameters are not defined.")

    @property
    def skew(self):
        """
        Distribution skewness

        Returns
        -------
        s : float
            distribution skewness
        """
        try:
            # zetac is the complementary Riemann zeta function (zeta function minus 1)
            # http://docs.scipy.org/doc/scipy/reference/generated/scipy.special.zetac.html
            s = -12. * np.sqrt(6.) * (1. + zetac(3)) / np.pi ** 3
            return s
        except TypeError:
            print("Distribution parameters are not defined.")

    def bootstrap(self, size=None, method='msm', N=100):
        """
        Parametric bootstrapping of source distribution

        Parameters
        ----------
        size : int
            bootstrap sample size. default equal to source sample size
        method : {'msm','lse','mle'}
            method of fit, optional
            'msm' = method of sample moments
            'lse' = least-square estimation
            'mle' = maximum likelihood estimation
        N : int
            number of bootstrap samples. default equal to 10

        Returns
        -------
        array-like
            m  - mean distribution parameters
        array_like
            cv - coefficient of variation of distribution parameter

        Notes
        -----
        In statistics, bootstrapping is a method for assigning measures of accuracy
        to sample estimates (variance,quantiles). This technique allows estimation of the
        sampling distribution  of almost any statistic using only very simple methods. Generally,
        it falls in the broader class of resampling methods. In this case a parametric model is fitted
        to the data, and samples of random numbers with the same size as the original data,
        are drawn from this fitted model. Then the quantity, or estimate, of interest is
        calculated from these data. This sampling process is repeated many times as for other
        bootstrap methods. If the results really matter, as many samples as is reasonable,
        given available computing power and time, should be used. Increasing the number of
        samples cannot increase the amount of information in the original data, it can only
        reduce the effects of random sampling errors which can arise from a bootstrap procedure itself.
        See [5] about bootstrapping.

        """
        options = {'msm': self._msm, 'lse': self._lse, 'mle': self._mle}
        assert method.lower() in options.keys(), "Method must be either %s" % (' or '.join(options.keys()))

        if size is None:
            assert self.data is not None, "Either size has to be specified or a sample has to be specified."
            size = np.size(self.data)

        i = 0
        par = np.zeros((N, 2))
        while (i < N):
            x = self.rnd(size=size)
            par[i, :] = options[method](x)
            i += 1

        m = par.mean(axis=0)
        cv = par.std(axis=0, ddof=1) / m

        return m, cv

    def cdf(self, x=None):
        """
        Cumulative distribution function (cumulative probability) for specified values x

        Parameters
        ----------
        x : array_like
            values

        Returns
        -------
        cdf : array
            cumulative probabilities for specified values x

        Notes
        -----
        A range of x values [location, location+3*std] are applied if x is not specified.
        """
        try:
            if x is None:
                x = np.linspace(self.loc, self.loc - 3. * self.std, 100)
            else:
                x = np.array(x)

            assert self.scale > 0., "The scale parameter must be larger than 0."

            z = (x - self.location) / self.scale
            p = 1. - np.exp(-np.exp(z))
            return p
        except TypeError:
            print("Distribution parameters are not defined")

    def fit(self, data=None, method='msm', verbose=False):
        """
        Determine distribution parameters by fit to sample.

        Parameters
        ----------
        data : array_like
            sample, optional
        method : {'msm','lse','mle'}
            method of fit, optional
            'msm' = method of sample moments
            'lse' = least-square estimation
            'mle' = maximum likelihood estimation
        verbose : bool
            turn on output of fitted parameters

        Notes
        -----
        If data is not input any data stored in object (self.data) will be used.

        """

        options = {'msm': msm, 'lse': lse, 'mle': mle}
        assert method.lower() in options.keys(), "Method must be either %s" % (' or '.join(options.keys()))

        if data is not None:
            # update sample data
            self.data = np.array(data).reshape(np.shape(data))  # make vector shaped

        try:
            self.location, self.scale = options[method](self.data)
            if verbose:
                print("Fitted parameters:\nlocation = %(location)5.3g\nscale = %(scale)5.3g" % self.__dict__)
        except TypeError:
            print("The sample data is not defined.")

    def fit_from_weibull_parameters(self, wa, wb, wc, n, verbose=False):
        """
        Calculate Gumbel distribution parameters from n independent Weibull distributed variables.

        Parameters
        ----------
        wa : float
            Weibull location parameter
        wb : float
            Weibull scale parameter
        wc : float
            Weibull shape parameter
        n : int
            Number independently distributed variables
        verbose : bool
            print fitted parameters

        Notes
        -----
        A warning is issued if Weibull shape parameter less than 1. In this case,
        the convergence towards asymptotic extreme value distribution is slow
        , and the asymptotic distribution will be non-conservative relative
        to the exact distribution. The asymptotic distribution is correct with Weibull
        shape equal to 1 and conservative with Weibull shape larger than 1.
        These deviations diminish with larger samples. See [1, p. 380].

        """
        # TODO:
        #  self.location = wa + wb * np.log(n) ** (1. / wc)
        #  self.scale = 1. / (wc / wb * np.log(n) ** ((wc - 1.) / wc))
        #  if verbose:
        #     print "Fitted parameters:\nlocation = %(location)5.3g\nscale = %(scale)5.3g" % self.__dict__
        raise NotImplementedError("Formula for deriving Gumbel minimum distribution parameter is not implemented.")

    def gp_plot(self, showfig=True, save=None):
        """
        Plot data on Gumbel paper (linearized scales))

        Parameters
        ----------
        showfig : bool
            show figure immediately on screen, default True
        save : filename
            save figure to file, default None

        """

        figure()

        x = np.sort(self.data)

        # sample
        z_data = -np.log(-np.log(self.ecdf))
        plot(x, z_data, 'ko', label='Data')

        # fit distributions
        a_msm, b_msm = msm(self.data)
        a_mle, b_mle = mle(self.data)
        a_lse, b_lse = lse(self.data)

        z_msm = (x - a_msm) / b_msm
        z_mle = (x - a_mle) / b_mle
        z_lse = (x - a_lse) / b_lse

        plot(x, z_msm, '-r', label='MSM')
        plot(x, z_mle, '--g', label='MLE')
        plot(x, z_lse, ':b', label='LSE')

        # plotting positions and plot configurations
        p = np.array([0.1, 0.2, 0.5, 0.7, 0.8, 0.9, 0.95, 0.99, 0.999, 0.9999])
        y = -np.log(-np.log(p))
        yticks(y, p)
        legend(loc='upper left')
        ylim(y[0], y[-1])
        xlabel('X')
        ylabel('Cumulative probability')
        grid(True)

        if showfig:
            show()
        elif save is not None:
            savefig(save)
        else:
            pass

    def invcdf(self, p=None):
        """
        Inverse cumulative distribution function for specified quantiles p

        Parameters
        ----------
        p : array_like
            quantiles (or. cumulative probabilities if you like)

        Returns
        -------
        x : array
            values corresponding to the specified quantiles

        Notes
        -----
        A range of quantiles from 0.001 to 0.999 are applied if quantiles are not specified
        """
        try:
            if p is None:
                p = np.linspace(0.001, 0.999, 100)
            else:
                p = np.array(p)

            assert self.scale > 0., "The scale parameter must be larger than 0."

            x = np.zeros(np.shape(p))

            x[p == 1.] = np.inf  # asymptotic
            x[(p < 0.) | (p > 1.)] = np.nan  # probabilities out of bounds
            z = (p >= 0.) & (p < 1.)  # valid quantile range
            x[z] = self.location + self.scale * np.log(-np.log(1.-p[z]))
            return x
        except TypeError:
            print("Distribution parameters are not defined")

    def pdf(self, x=None):
        """
        Probability density function for specified values x

        Parameters
        ----------
        x : array_like
            values

        Returns
        -------
        pdf : array
            probability density function for specified values x

        Notes
        -----
        A range of x values [location, location+3*std] are applied if x is not specified.
        """
        try:
            if x is None:
                x = np.linspace(self.loc, self.loc - 3. * self.std, 100)
            else:
                x = np.array(x)

            assert self.scale > 0., "The scale parameter must be larger than 0."

            z = (x - self.location) / self.scale
            p = (1. / self.scale) * np.exp(z - np.exp(z))
            return p
        except TypeError:
            print("Distribution parameters are not defined")

    def plot(self, showfig=True, save=None):
        """
        Plot data on regular scales

        Parameters
        ----------
        showfig : bool
            show figure immediately on screen, default True
        save : filename including suffix
            save figure to file, default None

        """

        figure()

        if self.data is not None:
            plot(np.sort(self.data), self.ecdf, 'ko', label='Data')

        y = np.linspace(0.001, 0.9999, 1000)
        x = self.invcdf(p=y)

        plot(x, y, '-r', label='Fitted')
        xlabel('X')
        ylabel('Cumulative probability')
        legend(loc='upper left')
        grid(True)

        if showfig:
            show()
        elif save is not None:
            savefig(save)
        else:
            pass

    def rnd(self, size=None, seed=None):
        """
        Draw random samples from probability distribution

        Parameters
        ----------
        size : int|numpy shape, optional
            sample size (default 1 random value is returned)
        seed : int
            seed for random number generator (default seed is random)

        Returns
        -------
        x : array
            random sample

        """

        if seed is not None:
            np.random.seed(seed)
        r = np.random.random_sample(size)
        x = self.invcdf(p=r)

        return x


def lse(x):
    """
    Fit distribution parameters to sample by method of least square fit to empirical cdf

    Parameters
    ----------
    x : array_like
        sample

    Notes
    -----
    Uses an approximate median rank estimate for the empirical cdf.
    """
    x = np.sort(x)
    f = empirical_cdf(x.size, kind='median')
    fp = lambda v, z: 1. - np.exp(-np.exp((z - v[0]) / v[1]))  # parametric Gumbel function
    e = lambda v, z, y: (fp(v, z) - y)  # error function to be minimized
    a0, b0 = msm(x)  # initial guess based on method of moments

    # least square fit
    p, cov, info, msg, ier = leastsq(e, [a0, b0], args=(x, f), full_output=1)

    return p[0], p[1]


def mle(x):
    """
    Fit distribution parameters to sample by maximum likelihood estimation

    Parameters
    ----------
    x : array_like
        sample

    Notes
    -----
    MLE equation set is given in 'Statistical Distributions' by Forbes et.al. (2010) and referred
    at [4]
    """

    def mle_eq(p, z):
        """
        MLE equation set

        Parameters
        ----------
        p : list
            distribution parameters a, b, c
        z : array_like
            data
        """
        loc, scale = p  # unpack parameters
        n = z.size

        out = [loc + scale * np.log(1. / n * np.sum(np.exp(z / scale))),
               z.mean() - np.sum(z * np.exp(z / scale)) / np.sum(np.exp(z / scale)) - scale]

        return out

    x = np.array(x)
    a0, b0 = msm(x)  # initial guess
    a, b = fsolve(mle_eq, [a0, b0], args=x)

    return a, b


def msm(x):
    """
    Fit distribution parameters to sample by method of sample moments

    Parameters
    ----------
    x : array_like
        sample

    Notes
    -----
    See description in [1] and [2].
    """
    x = np.array(x)

    b = np.sqrt(6.) * np.std(x, ddof=1) / np.pi  # using the unbiased sample standard deviation
    a = x.mean() + em() * b

    return a, b



#!/usr/bin/env python
# coding: utf-8
"""
Rainflow cycle counting algorithm according to ASTM E1049-85 (2011), section 5.4.4.
"""
from collections import deque
from itertools import chain

import numpy as np

# TODO: Evaluate from-to counting which stores the "orientation" of each cycle. Enables reconstruction of a time history


def reversals(series, endpoints=False):
    """
    A generator function which iterates over the reversals in the iterable *series*.

    Parameters
    ----------
    series : array_like
        data series
    endpoints : bool, optional
        If True, first and last points in `series` are included. Default is False.
        Note that in general, inclusion of end points is only relevant if the series passed is already an array of
        reversal points.

    Returns
    -------
    generator
        yields floats

    Notes
    -----
    Reversals are the points at which the first derivative on the series changes sign. The generator never yields
    the first and the last points in the series, unless `endpoints` is set to True (in which case they are always
    included).

    """
    series = iter(series)

    x_last, x = next(series), next(series)
    d_last = (x - x_last)

    if endpoints is True:
        yield x_last

    for x_next in series:
        if x_next == x:
            continue
        d_next = x_next - x
        if d_last * d_next < 0:
            yield x
        x_last, x = x, x_next
        d_last = d_next

    if endpoints is True:
        yield x


def cycles(series, endpoints=False):
    """
    Find full cycles and half-cycles range and mean value from `series` and count the number of occurrences using the
    Rainflow algorithm.

    Parameters
    ----------
    series : array_like
        data series
    endpoints : bool, optional
        If True, first and last points in `series` are included as cycle start/end points. This is convenient if the
        series given is already an array of reversal points. Default is False.

    Returns
    -------
    full : list
        full cycles (range and mean).
    half : list
        half cycles (range and mean).

    Notes
    -----
    The cycles are extracted from the iterable `series` according to section 5.4.4 in ASTM E1049 (2011).

    Examples
    --------
    Extract start and end points for all full and half cycles.

    >>> from qats.fatigue.rainflow import cycles
    >>> series = [0, -2, 1, -3, 5, -1, 3, -4, 4, -2, 0]
    >>> full, half = cycles(series)
    >>> full
    [(4, 1.0)]
    >>> half
    [(3, -0.5), (4, -1.0), (8, 1.0), (6, 1.0), (8, 0.0), (9, 0.5)]

    """
    points = deque()
    full, half = list(), list()

    for r in reversals(series, endpoints=endpoints):
        points.append(r)
        while len(points) >= 3:
            # Form ranges X and Y from the three most recent points
            x = abs(points[-2] - points[-1])
            y = abs(points[-3] - points[-2])        # cycle range
            m = 0.5 * (points[-2] + points[-3])     # cycle mean
            # fromto = (points[-2], points[-3])     # cycle start and end points (not currently in use)

            if x < y:
                # Read the next point
                break
            elif len(points) == 3:
                # Y contains the starting point
                # Count Y as half cycle and discard the first point
                half.append((y, m))
                points.popleft()
            else:
                # Count Y as one cycle and discard the peak and the valley of Y
                full.append((y, m))
                last = points.pop()
                points.pop()
                points.pop()
                points.append(last)
    else:
        # Count the remaining ranges as half cycles
        while len(points) > 1:
            half.append((abs(points[-2] - points[-1]), 0.5 * (points[-1] + points[-2])))
            points.pop()

    return full, half


def count_cycles(series, endpoints=False):
    """
    Count number of occurrences of cycle range and mean combinations.

    Parameters
    ----------
    series : array_like
        Data series.
    endpoints : bool, optional
        If True, first and last points in `series` are included as cycle start/end points. This is convenient if the
        series given is already an array of reversal points, but should in general not be used otherwise.
        Default is False.

    Returns
    -------
    cycles : np.ndarray
        Array of shape (n, 3) where `n` is number of cycle ranges. Each row consists of three values; the cycle
        range, mean and count. Counts are either 1.0 (for full cycles) or 0.5 (for half cycles).
        The array is sorted by increasing cycle range.

    Notes
    -----
    The cycles are extracted from the iterable series using the `cycles()` function.

    Since half cycles are counted as 0.5, the returned counts are not necessarily whole numbers.

    Examples
    --------
    Extract raw cycle range, mean and count:

    >>> from qats.fatigue.rainflow import count_cycles
    >>> series = [0, -2, 1, -3, 5, -1, 3, -4, 4, -2, 0]
    >>> count_cycles(series)
    array([[ 3. , -0.5,  0.5],
           [ 4. , -1. ,  0.5],
           [ 4. ,  1. ,  1. ],
           [ 6. ,  1. ,  0.5],
           [ 8. ,  0. ,  0.5],
           [ 8. ,  1. ,  0.5],
           [ 9. ,  0.5,  0.5]])

    The array may be unpacked into separate arrays of cycle range, mean and count as:

    >>> r, m, c = count_cycles(series).T

    The following will also work, but is slower than the example above:

    >>> r, m, c = zip(*count_cycles(series))

    See Also
    --------
    reversals, cycles

    """
    full, half = cycles(series, endpoints=endpoints)

    # number of cycles (full, half, total)
    nf = len(full)
    nh = len(half)
    n = nf + nh

    # initiate and populate array with cycle counts
    cycles_ = np.zeros((n, 3))
    cycles_[:, :2] = full + half  # full and half are lists
    cycles_[:nf, 2] = 1.0  # full cycles count 1.0
    cycles_[nf:, 2] = 0.5  # half cycles count 0.5

    # sort by increasing range, then mean
    cycles_ = _sort_cycles(cycles_, copy=False)

    return cycles_


def mesh(cycles, nr=100, nm=100):
    """
    Mesh range-mean distribution.

    Parameters
    ----------
    cycles : array_like
        Array of shape (n, 3). Columns should be: cycle range, cycle mean, count. See description of output
        from :func:`count_cycles`.
    nr : int, optional
        Number of equidistant bins for cycle ranges.
    nm : int, optional
        Number of equidistant bins for cycle means.

    Returns
    -------
    rangemesh : np.ndarray
        Cycle ranges meshgrid, shape: `(nm, nr)`.
    meanmesh : np.ndarray
        Cycle mean value meshgrid, shape: `(nm, nr)`.
    countmesh : np.ndarray
        Cycle count 2D histogram, shape: `(nm, nr)`.

    See Also
    --------
    numpy.histogram2d
    numpy.meshgrid

    Notes
    -----
    .. versionadded :: 4.7.0

    This function has been re-written for version 4.7.0. For versions <= 4.6.1, the mesh established was not correct.

    Shape of the returned arrays is consistent with :func:`numpy.meshgrid`: ``(nm, nr)``, i.e. number of rows is `nm` and
    number of columns is `nr`. This means that the array is transposed compared the output from
    :func:`numpy.histogram2d`, which is a 2D histogram of shape ``(nr, nm)``.

    The cycle count mesh is consistent with the cycles returned from :func:`rebin`, such that the sum of the cycle count
    mesh along each of its axes is equals the counts for cycles rebinned by 'mean' or 'range', respectively:

    >>> cycles = count_cycles(series)
    >>> _, _, cmesh = mesh(cycles, nr=200, nm=100)
    >>> cycles_rebinned_range = rebin(cycles, binby='range', n=200)
    >>> # sum of `cmesh` along cycle mean axis (constant cycle range) vs. counts from rebinned cycles
    >>> (cmesh.sum(axis=0) == cycles_rebinned_range[:, 2]).all()
    True
    >>> cmesh.sum(axis=0).shape
    (200,)
    >>> cmesh.sum(axis=1).shape
    (100,)


    Examples
    --------
    Rebin the cycle distribution onto a 10 x 15 mesh:

    >>> from qats.fatigue.rainflow import count_cycles, mesh
    >>> # (obtain some series from e.g. a simulation)
    >>> count_cycles(series)
    >>> rangemesh, meanmesh, countmesh = mesh(cycles, nr=15, nm=10)
    >>> countmesh.shape  # (same shape for all three arrays)
    (10, 15)

    The mesh returned is suitable for plotting with `matplotlib` 3D plots, for instance:

    >>> import matplotlib as mpl
    >>> import matplotlib.pyplot as plt
    >>> from mpl_toolkits.mplot3d import Axes3D
    >>> fig = plt.figure()
    >>> ax = fig.gca(projection='3d')
    >>> ax.plot_surface(rangemesh, meanmesh, countmesh, cmap=mpl.cm.coolwarm)
    >>> ax.set_xlabel('Cycle range')
    >>> ax.set_ylabel('Cycle mean')
    >>> ax.set_zlabel('Cycle count')
    >>> plt.show()

    """
    # unpack
    cycles = _toarray(cycles)  # ensure array, not list
    ranges, means, counts = cycles.T

    # create mesh
    maxrange = ranges.max()
    maxmean = means.max()
    minmean = means.min()

    # xyrange = ([0., maxrange], [minmean, maxmean])
    xyrange = ([0., maxrange], [minmean, maxmean])
    hist2d, r_edges, m_edges = np.histogram2d(ranges, means, bins=[nr, nm], range=xyrange, weights=counts)

    # 2D histogram from np.histogram2d must be transposed for consistency with np.meshgrid
    cmesh = hist2d.T

    # arrays of bin mid points (sizes `nr` and `rm`)
    rbins = 0.5 * (r_edges[:-1] + r_edges[1:])
    mbins = 0.5 * (m_edges[:-1] + m_edges[1:])

    # mesh grids
    rmesh, mmesh = np.meshgrid(rbins, mbins)

    return rmesh, mmesh, cmesh


def rebin(cycles, binby='range', n=None, w=None):
    """
    Rebin cycles in specified bins by range or mean value

    Parameters
    ----------
    cycles : array_like
        Array of shape (n, 3). Columns should be: cycle range, cycle mean, count. See description of output
        from `rainflow.count_cycles()`.
    binby : str, optional
        'range' - Rebin by cycle range (default)
        'mean'  - Rebin by cycle mean
    n : int, optional
        Number of equidistant bins for cycle ranges and cycle mean values.
    w : float, optional
        Width of equidistant bins for cycle ranges and cycle mean values. Overrides `n` if specified.

    Returns
    -------
    cycles : np.ndarray
        Array with shape `(nbins, 3)`, where `nbins` is number of bins. Columns are: cycle range, mean, count.

    Notes
    -----
    Cycles are gathered into a specific bin if the primary measure (range or mean) is within that bin's boundaries. The
    primary measure is represented by the bin mid point. The secondary measure (range or mean) is represented by its
    weighted average (weighted by number of occurrences) in the bin.

    Note that rebinning may shift a significant amount of the cycles to bins which midpoint differs notably from
    the original cycle value (range or mean). We advice to rebin for plotting, but when calculating e.g. fatigue damage
    we advice to use the raw unbinned values.

    .. versionadded :: 4.7.0

    If a cycle is at the edge between two bins, it is placed in the 'highest' bin. E.g. If bin edges are [0, 1, 2, 3],
    a cycle with value 1 will be counted in the bin [1, 2) - see documentation for np.histogram for details. This
    behaviour is contrary to version <= 4.6.1, for which the value 1 was counted in bin (0, 1]. The new way is slightly
    more conservative, since cycle ranges (or means) that end up on an edge is shifted towards a higher value.
    However; in most cases, there will be no difference at all since very few cycles coincide with bin edges.

    Examples
    --------
    Rebinning by range with bin width 1.0. Number of bins is then determined from the max cycle range.
    The second column (which is here the secondary measure, since we are binning by range) is the weighted average
    of the mean value for the cycles that fall within each bin:

    >>> from qats.fatigue.rainflow import count_cycles
    >>> series = [0, -2, 1, -3, 1, -1, 3, -2, 2, -2, 0, 1, -2, 3, -1, 2, -3, 0, -1, 0]
    >>> cycles = count_cycles(series)
    >>> rebin(cycles, binby='range', w=2.0)
    array([[ 1.  , -0.5 ,  0.5 ],
           [ 3.  , -0.25,  4.  ],
           [ 5.  ,  0.  ,  3.5 ]])

    See Also
    --------
    numpy.histogram
    count_cycles

    """
    if binby not in ('range', 'mean'):
        raise ValueError(f"Invalid option for `binby`: '{binby}'. Must be either 'range' or 'mean'")

    # unpack
    cycles = _toarray(cycles)  # ensure array, not list
    ranges, means, counts = cycles.T

    # rebin
    if binby == 'range':
        # establish bin edges
        bins = _create_bins(0., ranges.max(), n=n, w=w)
        # nbins = bins.size - 1

        # establish bin mid points (size -1 compared to array with bin edges)
        bin_primary = 0.5 * (bins[:-1] + bins[1:])

        # calculate sum of counts for each bin using np.histogram(). `range` has no effect since bins are given
        # explicitly, and is therefore passed as None
        bin_n, _ = np.histogram(ranges, bins=bins, range=None, weights=counts)

        # weighted average of cycle means is established in a similar way, however; dividing by total count needs to be
        # done afterwards. For bins with no count, the weighted average is set to nan.
        weights = counts * means
        bin_secondary, _ = np.histogram(ranges, bins=bins, range=None, weights=weights)
        bin_secondary[bin_n > 0] *= 1. / bin_n[bin_n > 0]
        bin_secondary[bin_n == 0] = np.nan

        return np.array([bin_primary, bin_secondary, bin_n]).T

    else:  # i.e. binby == 'mean'
        # create bins
        bins = _create_bins(means.min(), means.max(), n=n, w=w)
        # nbins = bins.size - 1

        # establish bin mid points (size -1 compared to array with bin edges)
        bin_primary = 0.5 * (bins[:-1] + bins[1:])

        # calculate sum of counts for each bin using np.histogram(). `range` has no effect since bins are given
        # explicitly, and is therefore passed as None
        bin_n, _ = np.histogram(means, bins=bins, range=None, weights=counts)

        # weighted average of cycle range is established in a similar way, however; dividing by total count needs to be
        # done afterwards. For bins with no count, the weighted average is set to nan.
        weights = counts * ranges
        bin_secondary, _ = np.histogram(means, bins=bins, range=None, weights=weights)
        bin_secondary[bin_n > 0] *= 1. / bin_n[bin_n > 0]
        bin_secondary[bin_n == 0] = np.nan

        return np.array([bin_secondary, bin_primary, bin_n]).T


def _create_bins(start, stop, n=None, w=None):
    """
    Create equidistant bins.

    Parameters
    ----------
    start : float
        Lowest bin value
    stop : float
        Largest bin value
    n : int, optional
        Number of equidistant bins
    w : float, optional
        Width of equidistant bins. Overrules `n` is specified.

    Returns
    -------
    np.ndarray
        Bins specified by bin boundaries e.g. first bin ranges from the first to the second value in `bins`,
        the next bin from the second value to the third, and so on.

    """
    if (not n) and (not w):
        raise ValueError('Specify either the number of bins `n` or the bin width `w`.')

    if w is not None:
        # crate bins with specified w
        return np.arange(start, stop + w, w)
    else:
        # create specified number of bins
        return np.linspace(start, stop, n + 1)


def _toarray(longlist) -> np.ndarray:
    """
    Convert (potentially long) list to 2D numpy array.

    Parameters
    ----------
    longlist : list or np.ndarray
        List that may be broadcast into 2D numpy array.

    Returns
    -------
    np.ndarray

    Notes
    -----
    Code is based on:
    https://stackoverflow.com/questions/17973507/why-is-converting-a-long-2d-list-to-numpy-array-so-slow

    Using ipython %timeit command, this function performed around 0.7s per loop for a list of length ~4500000
    (with 3 items in each row).

    The following alternatives performed at around 2.1s per loop:
    >>> # alternative 1
    >>> arr = np.array(longlist)
    >>> # alternative 2
    >>> arr = np.zeros((len(longlist), len(longlist[0])))
    >>> arr[:] = longlist

    ... and this alternative performed at around 2.5s per loop:
    >>> arr = np.array([[for row[i] in longlist] for i in range(len(longlist[0]))])

    """
    if isinstance(longlist, np.ndarray):
        # return quickly if input is already a numpy array
        return longlist
    flat = np.fromiter(chain.from_iterable(longlist), np.array(longlist[0][0]).dtype, -1)
    return flat.reshape((len(longlist), -1))


def _sort_cycles(arr, copy=False):
    """
    Sort cycles array wrt. 1st column then 2nd column.

    Parameters
    ----------
    arr : np.ndarray
        2D array of shape (n, 3).
    copy : bool, optional
        If True, return a copy instead of sorting the array inplace.
        Note that this increases time consumption, in particular for large arrays.

    Returns
    -------
    np.ndarray

    Raises
    ------
    IndexError
        If input array is 1D or if second dimension is less than 2.
    """
    if copy:
        # make a copy
        arr = np.array(arr)
    # using numpy.argsort(), ref. https://stackoverflow.com/a/38194077
    arr = arr[arr[:, 1].argsort()]  # first sort doesn't need to be stable
    arr = arr[arr[:, 0].argsort(kind='mergesort')]
    return arr




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:class:`Weibull` class and functions related to Weibull distribution.
"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import brentq, fsolve, leastsq
from scipy.special import binom, gamma

from ..signal import find_maxima
from .empirical import empirical_cdf

# todo: build documentation and check that docstrings behave as intended
# todo: create examples
# todo: consider replacing latex with python code or pseudo code (ref. link below)
# link: https://github.com/numpy/numpy/blob/master/doc/HOWTO_DOCUMENT.rst.txt#other-points-to-keep-in-mind


class Weibull(object):
    """
    The Weibull class offers miscellaneous functions for working with the Weibull
    distribution, defined as (cumulative distribution function)::

        F(x) = 1 - exp{-[(x-a)/b]^c}

    where `a` is location parameter, `b` is scale parameter and `c` is shape parameter.


    Parameters
    ----------
    loc: float
        Weibull location parameter.
    scale: float
        Weibull scale parameter.
    shape: float
        Weibull shape parameter.
    data: array_like, optional
        Sample data, used to establish empirical cdf and is included in plots.
        To fit the Weibull distribution to the sample data, use :meth:`Weibull.fit()`.

    Attributes
    ----------
    loc: float
        Weibull location parameter.
    scale: float
        Weibull scale parameter.
    shape: float
        Weibull shape parameter.
    data: array_like
        Sample data. Exists only if distribution parameters are estimated from a sample.

    Notes
    -----
    For a Weibull 2-parameter distribution, specify location parameter 0 (zero).


    Examples
    --------
    To initiate an instance based on parameters, use:

    >>> from qats.stats.weibull import Weibull
    >>> weib = Weibull(loc, scale, shape)

    If you need to establish a Weibull instance based on a sample data set, use:

    >>> weib = Weibull.fit(data, method='pwm')


    References
    ----------
    1. Moment estimators for Weibull parameters and their asymptotic efficiencies, Waloddi Weibull, April 1969, Lausanne Switzerland, Technical report AFML-TR-69-135

    2. Continuous univariate distributions, Volume 1, N.L.Johnson, S.Kotz and N.Balakrishnan, 1994,  John Wiley and sons inc.

    3. `weibull.com <http://www.weibull.com/hotwire/issue15/relbasics15.htm>`_, About location parameter

    4. `Plotting positions <http://en.wikipedia.org/wiki/Q%E2%80%93Q_plot>`_, About plotting positions

    5. `Bootstrapping <https://en.wikipedia.org/wiki/Bootstrapping_(statistics)>`_, Bootstrapping statistics

    6. Estimation of the generalized extreme value distribution by the method of probability weighted moments, Hosking, J. R. M., Wallis, J. R. and Wood, E. F., 1985, Technometrics, 27, pp. 251-261

    7. Estimating the three-parameter Weibull distribution by the method of probability weighted moments with application to medical survival data, Bortolucci, A. A. et.al.

    8. Theory and derivation for Weibull parameter probability weighted moment estimators, Grender, J.M., Dell, T.R., Reich, R.M., 1991 United Sates Department of Agriculture

    9. Probability weighted moments, Greenwood, J. A.; Landwehr, J.M.; Matalas, N.C.; Wallis, J.R., 1979, Water Resources Research. 15(5): 1049-1054.

    10. Probability weighted moments compared with some traditional techniques in estimating gumbel parameters and quantiles., Landwehr, J.M.; Matalas, N.C.; Wallis, J.R., 1979., Water Resources Research. 15(5): 1063-1064.

    """

    def __init__(self, loc, scale, shape, data=None):
        self.loc = loc
        self.scale = scale
        self.shape = shape

        if data is not None:
            self.data = np.array(data)
        else:
            self.data = None

    @property
    def cov(self):
        """
        Distribution coefficient of variation (C.O.V.)

        Returns
        -------
        float
            distribution c.o.v.
        """
        return self.std / self.mean

    @property
    def ecdf(self):
        """
        Empirical cumulative distribution function associated with the sample.

        Returns
        -------
        array
            Empirical cumulative distribution function.

        Notes
        -----
        A mean rank method is chosen to approximate the mean of the distribution [2].

        The empirical cdf is also used as plotting positions when plotting the sample
        on probability paper.

        """
        p = empirical_cdf(self.data.size, kind='mean')
        return p

    @property
    def kurt(self):
        """
        Distribution kurtosis.

        Returns
        -------
        float
            distribution kurtosis
        """
        r1 = gamma(1. + 1. / self.shape)
        r2 = gamma(1. + 2. / self.shape)
        r3 = gamma(1. + 3. / self.shape)
        r4 = gamma(1. + 4. / self.shape)

        k = (r4 + r1 * (-4.0 * r3 + r1 * (6.0 * r2 - 3.0 * r1 ** 2))) / (r2 - r1 * r1)
        return k

    @property
    def mean(self):
        """
        Distribution mean value

        Returns
        -------
        float
            distribution mean value
        """
        return self.loc + self.scale * gamma(1. + 1. / self.shape)

    @property
    def mse(self):
        """
        Mean squared error of fitted cumulative distribution (a,b,c) and empirical distribution

        Returns
        -------
        float
            mean squared error

        """
        e = np.sum((self.ecdf - self.cdf(np.sort(self.data))) ** 2.)/self.data.size
        return e

    @property
    def params(self):
        """
        Distribution parameters.

        Returns
        -------
        tuple
            Distribution parameters: (loc, scale, shape).
        """
        return self.loc, self.scale, self.shape

    @property
    def skew(self):
        """
        Distribution skewness

        Returns
        -------
        float
            distribution skewness
        """
        r1 = gamma(1. + 1. / self.shape)
        r2 = gamma(1. + 2. / self.shape)
        r3 = gamma(1. + 3. / self.shape)
        s = (r3 + r1 * (2. * r1 ** 2. - 3. * r2)) / (r2 - r1 ** 2.) ** 1.5
        return s

    @property
    def std(self):
        """
        Distribution standard deviation

        Returns
        -------
        float
            distribution standard deviation
        """
        r1 = gamma(1. + 1. / self.shape)
        r2 = gamma(1. + 2. / self.shape)
        s = self.scale * np.sqrt(r2 - r1 ** 2.)
        return s

    def gumbel_parameters(self, n=None):
        """
        Calculate parameters of the asymptotic Gumbel extreme value distribution (Type 1) for the extreme value of N
        independent, Weibull distributed variables.

        Parameters
        ----------
        n : int
            number of independent weibull distributed variables, default equal to number of peaks (self.data.size)

        Returns
        -------
        tuple
            Gumbel location and scale parameters


        See Also
        --------
        qats.stats.weibull.weibull2gumbel


        Notes
        -----
        If the sample `x` is based on lets say a 30-hour simulation but you seek an estimate of the e.g. 3-hour extreme
        value then `n` should be calculated as the nearest integer to::

            n = (3 / 30) * nx

        where `nx` is the total number of maxima during 30 hour.


        References
        ----------
        1. Bury, K.V. (1975), "Statistical models in applied science"

        """
        if n is None:
            n = self.data.size
        gloc, gscale = weibull2gumbel(self.loc, self.scale, self.shape, n)

        return gloc, gscale

    def cdf(self, x=None):
        """
        Cumulative distribution function (cumulative probability) for specified values x

        Parameters
        ----------
        x : array_like
            values

        Returns
        -------
        array
            cumulative probabilities for specified values x

        Notes
        -----
        A range of x values are applied if x is not specified.
        """
        if x is None:
            x = np.linspace(self.loc, self.loc + 3. * self.std, 100)
        else:
            x = np.array(x)

        assert np.all(x >= self.loc), "The location parameter must be less than all items in data set"

        p = 1. - np.exp(-((x - self.loc) / self.scale) ** self.shape)
        return p

    @classmethod
    def fit(cls, data, method='msm', verbose=False):
        """
        Establish Weibull class instance by fit to sample.

        Parameters
        ----------
        data : array_like
            Sample.
        method : str, optional
            Method of fit. Available options:

            - ``msm`` = method of sample moments (default)
            - ``lse`` = least-square estimation
            - ``mle`` = maximum likelihood estimation
            - ``pwm`` = probability weighted moments
            - ``pwm2`` = probability weighted moments, 2-parameter distribution

        verbose : bool
            If True, fitted parameters are printed to screen.

        Returns
        -------
        Weibull
            Weibull class instance

        See Also
        --------
        Weibull.fromsignal

        Examples
        --------
        Assuming `data` is a sample array/list:

        >>> from qats.stats.weibull import Weibull
        >>> weib = Weibull.fit(data, method="msm")

        """
        options = {'msm': msm, 'lse': lse, 'mle': mle, 'pwm': pwm, 'pwm2': pwm2}
        assert method.lower() in options.keys(), "Method must be either %s" % (' or '.join(options.keys()))

        data = np.array(data)  # ensure numpy array

        params = options[method](data)
        if len(params) == 2:  # 2-parameter distribution
            loc = 0
            scale, shape = params
        else:                 # 3-parameter distribution
            loc, scale, shape = params

        if verbose:
            print("Fitted parameters:\n"
                  "location = %5.3g\n"
                  "scale = %5.3g\n"
                  "shape = %5.3g" % (loc, scale, shape)
                  )

        weib = cls(loc, scale, shape, data=data)
        return weib

    @classmethod
    def fromsignal(cls, x, method='msm', verbose=False):
        """
        Establish Weibull class instance by fit to global maxima from time series signal.

        Parameters
        ----------
        x : array_like
            Time series signal.
        method : str, optional
            Method of fit. See :meth:`Weibull.fit()` for description of options.
        verbose : bool, optional
            If True, fitted parameters are printed to screen.

        Returns
        -------
        Weibull
            Class instance.

        See Also
        --------
        Weibull.fit
        qats.stats.find_maxima

        Examples
        --------
        Assuming `x` is a time series signal:

        >>> from qats.stats.weibull import Weibull
        >>> weib = Weibull.fromsignal(x, method='msm')

        Note that the example above is equivalent to:

        >>> from qats.signal import find_maxima
        >>> sample, _ = find_maxima(x, local=False, threshold=None, up=True)
        >>> weib = Weibull.fit(sample, method='msm')
        """
        sample, _ = find_maxima(x, local=False, threshold=None, up=True)
        return Weibull.fit(sample, method=method, verbose=verbose)

    def invcdf(self, p=None):
        """
        Inverse cumulative distribution function for specified quantiles p

        Parameters
        ----------
        p : array_like
            quantiles (or. cumulative probabilities if you like)

        Returns
        -------
        array
            values corresponding to the specified quantiles

        Notes
        -----
        A range of quantiles from 0 to 1 are applied if quantiles are not specified

        """
        if p is None:
            p = np.linspace(0.001, 0.999, 100)
        else:
            p = np.array(p)

        x = np.zeros(np.shape(p))

        x[p == 1.] = np.inf  # asymptotic
        x[(p < 0.) | (p > 1.)] = np.nan  # probabilities out of bounds
        x[(p >= 0.) & (p < 1.)] = self.loc + self.scale * \
                                  (-np.log(1. - p[(p >= 0.) & (p < 1.)])) ** (1. / self.shape)
        return x

    def pdf(self, x=None):
        """
        Probability density function for specified values x

        Parameters
        ----------
        x : array_like
            values

        Returns
        -------
        array
            probability density function for specified values x

        Notes
        -----
        A range of x values are applied if x is not specified.
        """
        if x is None:
            x = np.linspace(self.loc, self.loc + 3. * self.std, 100)
        else:
            x = np.array(x)

        assert np.all(x >= self.loc), "The location parameter must be less than all items in data set"

        p = self.shape / self.scale * ((x - self.loc) / self.scale) ** (self.shape - 1.) * np.exp(
            -((x - self.loc) / self.scale) ** self.shape)
        return p

    def plot(self, filename=None):
        """
        Plot data on regular scales

        Parameters
        ----------
        filename : str, optional
            Save plot as `filename`, default is to show plot on screen

        Examples
        --------
        Plot distribution and show the figure

        >>> from qats.stats.weibull import Weibull
        >>> distribution = Weibull(100., 15., 2.5)
        >>> distribution.plot()

        Plot distribution and save the figure as png

        >>> from qats.stats.weibull import Weibull
        >>> distribution = Weibull(100., 15., 2.5)
        >>> distribution.plot(filename="plot.png")

        """
        plt.figure()

        # plot data if that exist
        if self.data is not None:
            plt.plot(np.sort(self.data), self.ecdf, 'ko', label='Data')

        # plot fitted/specified distribution
        y = np.linspace(0.001, 0.9999, 1000)
        x = self.invcdf(p=y)
        plt.plot(x, y, '-r', label='Fitted')
        plt.xlabel('X')
        plt.ylabel('Cumulative probability')
        plt.legend(loc='upper left')
        plt.grid(True)

        # show figure
        if filename is not None:
            plt.savefig(filename)
        else:
            plt.show()

    def plot_linear(self, filename=None):
        """
        Plot data on Weibull paper (linearized scales))

        Parameters
        ----------
        filename : str, optional
            Save plot as `filename`, default is to show plot on screen

        Examples
        --------
        Plot distribution and show the figure

        >>> from qats.stats.weibull import Weibull
        >>> distribution = Weibull(100., 15., 2.5)
        >>> distribution.plot_linear()

        Plot distribution and save the figure as png

        >>> from qats.stats.weibull import Weibull
        >>> distribution = Weibull(100., 15., 2.5)
        >>> distribution.plot_linear(filename="plot.png")

        References
        ----------
        1. `Continuous univariate distributions, Volume 1`, N.L.Johnson, S.Kotz and N.Balakrishnan, 1994, \
            John Wiley and sons inc.
        """
        plt.figure()

        # plot data if that exists
        if self.data is not None:
            x = np.log(np.sort(self.data - self.loc))
            y = np.log(-np.log(1. - self.ecdf))
            plt.plot(x, y, 'ko', label='Data')

        # plot fitted distribution
        yy = np.array([0.1, 0.2, 0.5, 0.7, 0.8, 0.9, 0.95, 0.99, 0.999, 0.9999])
        y = np.log(-np.log(1. - yy))
        x = np.log(self.invcdf(p=yy) - self.loc)
        plt.plot(x, y, label='Fitted')

        plt.xlim(x[0], x[-1])
        plt.ylim(y[0], y[-1])
        plt.xlabel('X')
        plt.ylabel('Cumulative probability')
        plt.yticks(y, yy)
        xloc = np.linspace(x[0], x[-1], 5)
        xlab = np.around(np.exp(xloc) + self.loc, decimals=2)
        plt.xticks(xloc, xlab)
        plt.legend(loc='upper left')
        plt.grid(True)

        if filename is not None:
            plt.savefig(filename)
        else:
            plt.show()

        # close figure
        plt.close()

    def rnd(self, size=None, seed=None):
        """
        Draw random samples from probability distribution

        Parameters
        ----------
        size : int|numpy shape, optional
            sample size (default 1 random value is returned)
        seed : int
            seed for random number generator (default seed is random)

        Returns
        -------
        array
            random sample

        """

        if seed is not None:
            np.random.seed(seed)
        r = np.random.random_sample(size)
        x = self.invcdf(p=r)

        return x


def bootstrap(loc, scale, shape, size, repetitions, method='pwm'):
    """
    Quantify mean and coefficient of variation of Weibull distribution parameters using parametric bootstrapping

    Parameters
    ----------
    loc : float
        Source distribution location parameter
    scale : float
        Source distribution scale parameter
    shape : float
        Source distribution shape parameter
    size : int
        Size of bootstrapped sample
    method : str, optional
        Method of fit. Available options:

        - ``msm`` = method of sample moments
        - ``lse`` = least-square estimation
        - ``mle`` = maximum likelihood estimation
        - ``pwm`` = probability weighted moments (default)

    repetitions : int, optional
        Number of bootstrap samples. default equal to 100

    Returns
    -------
    array
        Mean distribution parameters
    array
        Coefficient of variation of distribution parameter

    Notes
    -----
    In statistics, bootstrapping is a method for assigning measures of accuracy
    to sample estimates (variance,quantiles). This technique allows estimation of the
    sampling distribution  of almost any statistic using only very simple methods. Generally,
    it falls in the broader class of resampling methods. In this case a parametric model is fitted
    to the data, and samples of random numbers with the same size as the original data,
    are drawn from this fitted model. Then the quantity, or estimate, of interest is
    calculated from these data. This sampling process is repeated many times as for other
    bootstrap methods. If the results really matter, as many samples as is reasonable,
    given available computing power and time, should be used. Increasing the number of
    samples cannot increase the amount of information in the original data, it can only
    reduce the effects of random sampling errors which can arise from a bootstrap procedure itself.
    See [5] about bootstrapping.

    Examples
    --------
    To quantify the uncertainty (coefficient of variation) of a Weibull distribution fitted to a sample with 5 values
    (using 100 repetition):

    >>> from qats.stats.weibull import bootstrap
    >>> m, cv = bootstrap(10., 5., 2.5, 5, 100)

    """
    options = {'msm': msm, 'lse': lse, 'pwm': pwm, 'mle': mle}
    assert method.lower() in options.keys(), "Method must be either %s" % (' or '.join(options.keys()))

    # initiate distribution
    distribution = Weibull(loc, scale, shape)

    i = 0
    par = np.zeros((repetitions, 2))
    while i < repetitions:
        # sample random numbers from source distribution
        x = distribution.rnd(size=size)

        # estimate distribution parameters to random sample
        par[i, :] = options[method](x)
        i += 1

    # calculate mean value and coefficient of variation of the distribution parameters
    m = par.mean(axis=0)
    cv = par.std(axis=0, ddof=1) / m

    return m, cv


def lse(x, threshold: float = None):
    """
    Fit Weibull distribution parameters to sample by method of least square fit to empirical cdf.

    Parameters
    ----------
    x : array_like
        sample data
    threshold : float, optional
        Fit distribution to data points above this threshold. The threshold is defined as value <0, 1> in the
        empirical CDF. So with threshold=0.87 the distribution is fitted to the values exceeding the 0.87-quantile of
        the empirical cumulative distribution function.

    Returns
    -------
    tuple (floats)
        Distribution parameters ``(loc, scale, shape)``.

    Notes
    -----
    Uses what are known as (approximate) mean rank estimates for the empirical cdf.
    """
    x = np.sort(x)                          # sorted dataset
    f = empirical_cdf(x.size, kind='mean')  # mean rank empirical cdf according to Weibull [1]
    weights = np.ones(np.shape(x))          # by default include all data points
    if threshold is not None:
        # exclude data points below threshold (fit only to the tail)
        weights[f < threshold] = 0.

    # define error function
    # v: distribution parameters
    # z: data points
    # y: empirical cdf defined at data points x (z)
    # w: weights to exclude data points below threshold if tail fitting is requested
    fp = lambda v, z: 1. - np.exp(-((z - v[0]) / v[1]) ** v[2])  # parametric weibull function
    e = lambda v, z, y, w: w * (fp(v, z) - y)  # error function to be minimized
    a0, b0, c0 = msm(x)  # initial guess based on method of moments

    # least square fit (x corresponds to z in lambdas, f corresponds to y and w corresponds to weights)
    p, cov, info, msg, ier = leastsq(e, np.asarray([a0, b0, c0]), args=(x, f, weights), full_output=True)

    return p[0], p[1], p[2]


def mle(x):
    """
    Fit Weibull distribution parameters to sample by maximum likelihood estimation

    Parameters
    ----------
    x : array_like
        sample data

    Returns
    -------
    tuple (floats)
        Distribution parameters ``(loc, scale, shape)``.

    """

    def mle_eq(p, z):
        """
        MLE equation set according to [2, p. 656 eq. 21.73-21.75]

        Parameters
        ----------
        p : array_like
            distribution parameters a, b, c
        z : array_like
            data sample

        Returns
        -------
        tuple (floats)
            Estimates of distribution parameters ``(loc, scale, shape)``.
        """
        loc, scale, shape = p  # unpack parameters
        n = z.size

        f1 = n / shape + np.sum(np.log((z - loc) / scale)) - np.sum(
            ((z - loc) / scale) ** shape * np.log((z - loc) / scale))
        f2 = -n * shape / scale + shape / scale * np.sum(((z - loc) / scale) ** shape)
        f3 = -(shape - 1.) * np.sum(1. / (z - loc)) + shape / scale * np.sum(((z - loc) / scale) ** (shape - 1.))

        return f1, f2, f3

    x = np.array(x)
    a0, b0, c0 = msm(x)  # initial guess
    a, b, c = fsolve(mle_eq, np.asarray([a0, b0, c0]), args=x)

    return a, b, c


def mlj(sample, l, j):
    """
    Probability weighted moment `Mljk` of observation order l, order of cdf j, with emphasize on the
    right/upper tail (k=0).

    Parameters
    ----------
    sample : array
        Sample.
    l : int
        Order of observation (sample).
    j : int
        Order of cumulative distribution function.

    Returns
    -------
    float
        Probability weighted moment

    Notes
    -----
    The probability weighted moment `Mljk` is defined by Greenwood and others (1979) as::

        M_{l,j,k} = E[X^l * F^j * (1-F)^k]

    , where `X(F)` is the inverse form of the distribution and `F` is the cumulative distribution function.
    When `j=k=0` and `l` is a non-negative integer, then `M_{l,0,0}` represents the conventional moment of order `l`
    about the origin.

    PWMs can be applied either when the small observations are more important than the large observations (k=0), as in
    strength properties of materials, or when the large observations should have more influence than the smaller
    observations (k=0) as with three diameter distribution modelling. Here we have chosen the latter and derived
    unbiased estimators for moments `M_{l,j,0}(k=0)`, see eq. 32 in [8]::

        M_{l,j,0} = (1 / n) * sum(x[i]^l * binom(i-1, j) / binom(n-1, j))

    where `i` is a counter from `j+1` to `n` and `binom()` is the binomial coefficient.

    """
    # todo: include pseudo-code (or Sphinx-friendly LaTex code) for M_{l,j,0} as included below
    '''
    .. math:: M_{l,j,k} = E[X^l F^j (1-F)^k]
    .. math:: M_{l,j,0} = \frac{1}{n}\sum_{i=j+1}^{n}{X_{(i)}^l\frac{\binom{i-1}{j}}{\binom{n-1}{j}}}
    '''
    n = float(sample.size)
    xi = np.sort(sample)[j:]  # (j+1)th subsample of sorted sample
    ii = np.arange(j + 1, n + 1)

    # ratio of binomial coefficients, probability that xi is the largest observation in a subsample of size (j+1)
    # drawn from n.
    bc = np.array([binom(i - 1., j) / binom(n - 1., j) for i in ii])

    return (1. / n) * np.sum(xi ** l * bc)  # eq. 32 in [8]


def msm(x):
    """
    Fit Weibull distribution parameters to sample by method of sample moments

    Parameters
    ----------
    x : array_like
        sample data

    Returns
    -------
    floats
        The `loc`, `scale` and `shape` distribution parameters

    Notes
    -----
    See description in [1].
    """
    x = np.array(x)
    a1 = x.mean()  # first sample raw moment aka. mean
    m2 = x.var()  # second sample central moment aka. variance
    m3 = np.mean((x - a1) ** 3.)  # third sample central moment
    c1 = m3 / m2 ** (3. / 2.)  # coefficient of sample skewness

    # solve for shape parameter using root search, see [1, eq. 68 on p.11]
    def f(shape, shape0):
        eq = (gamma((shape + 3.) / shape) - 3. * gamma((shape + 1.) / shape) * gamma((shape + 2.) / shape) +
              2 * gamma((shape + 1.) / shape) ** 3.) / (gamma((shape + 2.) / shape) -
                                                        gamma((shape + 1.) / shape) ** 2.) ** (3. / 2.) - shape0
        return eq

    c = brentq(f, args=(c1,), a=0.1, b=1000., maxiter=1000)

    # calculate location and scale parameters, see [1, eq. 67 on p. 11]
    g1 = gamma((c + 1.) / c)
    g2 = gamma((c + 2.) / c)
    b = np.sqrt(m2 / (g2 - g1 ** 2.))
    a = a1 - g1 * b

    return a, b, c


def plot_fit(x: np.ndarray, params: tuple, path: str = None):
    """
    Plot data sample versus empirical and fitted cumulative distribution function on linearized Weibull scales

    Parameters
    ----------
    x : array_like
        Data sample
    params : tuple
        location, scale and shape parameter of the Weibull distribution
    path : str, optional
        Save figure to file instead of displaying it.

    """
    # unpack distribution parameters
    a, b, c = params

    # normalize data and empirical CDF
    x = np.sort(x)
    x_norm = np.log(x - a)
    y_norm = np.log(-np.log(1. - (np.arange(x.size) + 1.) / (x.size + 1.)))

    # labels and tick positions for weibull paper plot
    p = np.array([0.2, 0.5, 0.7, 0.8, 0.9, 0.95, 0.99, 0.999, 0.9999])
    p_norm = np.log(-np.log(1. - p))

    # normalize fitted distribution
    q_norm = np.log(b * (-np.log(1. - p)) ** (1. / c))

    # plot
    plt.figure()
    plt.plot(x_norm, y_norm, "ko", label="Data")
    plt.plot(q_norm, p_norm, label="Fit")
    plt.yticks(p_norm, p)
    xticks = np.linspace(q_norm[0], q_norm[-1], 5)
    xlabel = np.around(np.exp(xticks) + a, decimals=2)
    plt.xticks(xticks, xlabel)
    plt.title(f"a={a:5.3f}, b={b:5.3f}, c={c:5.3f}")
    plt.xlabel('X')
    plt.ylabel('Cumulative probability')
    plt.legend(loc='upper left')
    plt.grid(True)
    if path is not None:
        plt.savefig(path)
    else:
        plt.show()


def pwm(x):
    """
    Fit distribution parameters to sample by method of probability weighted
    moments

    Parameters
    ----------
    x : array_like
        sample data

    Returns
    -------
    floats
        The `loc`, `scale` and `shape` distribution parameters

    Notes
    -----
    Details on probability weighted moments are provided in [8].

    See Also
    --------
    mlj()

    """
    # probability weighted moments of order 0,1,2,3
    m100 = mlj(x, 1, 0)
    m110 = mlj(x, 1, 1)
    m120 = mlj(x, 1, 2)
    m130 = mlj(x, 1, 3)

    # parameter estimates
    c = np.log(2.) / np.log((2. * m110 - m100) / (2. * (5. * m110 - m100 - 6. * m120 + 2. * m130)))
    a = 4. * (m100 * (3. * m120 - m130 - m110) - m110 ** 2.) / (m100 - 8. * m110 + 12. * m120 - 4. * m130)
    b = (m100 - a) / gamma(1. + 1. / c)

    return a, b, c


def pwm2(x):
    """
    Fit distribution parameters to sample by method of probability weighted moments assuming the location parameter
    is zero.

    Parameters
    ----------
    x : array_like
        sample data

    Returns
    -------
    floats
        The `scale` and `shape` distribution parameters

    Notes
    -----
    Details on probability weighted moments are provided on p.14-15 in [8]. Note that only the scale and parameters are
    estimated, the location parameter is assumed zero.

    See Also
    --------
    mlj(), pwm()

    """
    # probability weighted moments of order 0,1
    m100 = mlj(x, 1, 0)
    m110 = mlj(x, 1, 1)

    # parameter estimates
    c = np.log(2.) / np.log(m100 / (2.*(m100 - m110)))
    b = m100 / gamma(1. + 1. / c)

    return b, c


def weibull2gumbel(loc, scale, shape, n):
    """
    Calculate parameters of the asymptotic Gumbel extreme value distribution (Type 1) for the extreme value of N
    independent,Weibull distributed variables.

    Parameters
    ----------
    loc : float
        Weibull distribution location parameter
    scale : float
        Weibull distribution scale parameter
    shape : float
        Weibull distribution shape parameter
    n : int
        Number of independent weibull distributed variables

    Returns
    -------
    tuple
        Gumbel location and scale parameters

    Notes
    -----
    If the sample `x` is based on lets say a 30-hour simulation but you seek
    an estimate of the e.g. 3-hour extreme value then `n` should be calculated as
    the nearest integer to::

        n = (3 / 30) * nx

    where `nx` is the total number of maxima during 30 hour.

    References
    ----------
    1. Bury, K.V. (1975), "Statistical models in applied science"

    """
    gloc = loc + scale * np.log(n) ** (1. / shape)
    gscale = 1. / (shape / scale * np.log(n) ** ((shape - 1.) / shape))

    return gloc, gscale
